from typing import Any, List
import re
from openai import AsyncOpenAI

from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.language_models import LanguageModel


class OpenAIService(InferenceServiceABC):
    """OpenAI service class."""

    _inference_service_ = "openai"
    _env_key_name_ = "OPENAI_API_KEY"

    # TODO: Make this a coop call
    model_exclude_list = [
        "whisper-1",
        "davinci-002",
        "dall-e-2",
        "tts-1-hd-1106",
        "tts-1-hd",
        "dall-e-3",
        "tts-1",
        "babbage-002",
        "tts-1-1106",
        "text-embedding-3-large",
        "text-embedding-3-small",
        "text-embedding-ada-002",
        "ft:davinci-002:mit-horton-lab::8OfuHgoo",
    ]
    _models_list_cache: List[str] = []

    @classmethod
    def available(cls) -> List[str]:
        from openai import OpenAI

        if not cls._models_list_cache:
            try:
                client = OpenAI()
                cls._models_list_cache = [
                    m.id
                    for m in client.models.list()
                    if m.id not in cls.model_exclude_list
                ]
            except Exception as e:
                print(
                    f"Error retrieving models: {e}. /nSee instructions about storing your API keys: https://docs.expectedparrot.com/en/latest/api_keys.html"
                )
                cls._models_list_cache = [
                    "gpt-3.5-turbo",
                    "gpt-4-1106-preview",
                    "gpt-4",
                ]  # Fallback list
        return cls._models_list_cache

    @classmethod
    def create_model(cls, model_name, model_class_name=None) -> LanguageModel:
        if model_class_name is None:
            model_class_name = cls.to_class_name(model_name)

        class LLM(LanguageModel):
            """
            Child class of LanguageModel for interacting with OpenAI models
            """

            _inference_service_ = cls._inference_service_
            _model_ = model_name
            _parameters_ = {
                "temperature": 0.5,
                "max_tokens": 1000,
                "top_p": 1,
                "frequency_penalty": 0,
                "presence_penalty": 0,
                "logprobs": False,
                "top_logprobs": 3,
            }

            @classmethod
            def available(cls) -> list[str]:
                client = openai.OpenAI()
                return client.models.list()

            def get_headers(self) -> dict[str, Any]:
                from openai import OpenAI

                client = OpenAI()
                response = client.chat.completions.with_raw_response.create(
                    messages=[
                        {
                            "role": "user",
                            "content": "Say this is a test",
                        }
                    ],
                    model=self.model,
                )
                return dict(response.headers)

            def get_rate_limits(self) -> dict[str, Any]:
                try:
                    headers = self.get_headers()
                except Exception as e:
                    return {
                        "rpm": 10_000,
                        "tpm": 2_000_000,
                    }
                else:
                    return {
                        "rpm": int(headers["x-ratelimit-limit-requests"]),
                        "tpm": int(headers["x-ratelimit-limit-tokens"]),
                    }

            async def async_execute_model_call(
                self,
                user_prompt: str,
                system_prompt: str = "",
                encoded_image=None,
            ) -> dict[str, Any]:
                """Calls the OpenAI API and returns the API response."""
                content = [{"type": "text", "text": user_prompt}]
                if encoded_image:
                    content.append(
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{encoded_image}"
                            },
                        }
                    )
                self.client = AsyncOpenAI()
                response = await self.client.chat.completions.create(
                    model=self.model,
                    messages=[
                        {"role": "system", "content": system_prompt},
                        {"role": "user", "content": content},
                    ],
                    temperature=self.temperature,
                    max_tokens=self.max_tokens,
                    top_p=self.top_p,
                    frequency_penalty=self.frequency_penalty,
                    presence_penalty=self.presence_penalty,
                    logprobs=self.logprobs,
                    top_logprobs=self.top_logprobs if self.logprobs else None,
                )
                return response.model_dump()

            @staticmethod
            def parse_response(raw_response: dict[str, Any]) -> str:
                """Parses the API response and returns the response text."""
                try:
                    response = raw_response["choices"][0]["message"]["content"]
                except KeyError:
                    print("Tried to parse response but failed:")
                    print(raw_response)
                pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
                match = re.match(pattern, response, re.DOTALL)
                if match:
                    return match.group(1)
                else:
                    return response

        LLM.__name__ = "LanguageModel"

        return LLM



from edsl.jobs.tokens.TokenUsage import TokenUsage
from edsl.enums import TokenPricing


class InterviewTokenUsage:
    def __init__(
        self, new_token_usage: TokenUsage = None, cached_token_usage: TokenUsage = None
    ):
        self.new_token_usage = new_token_usage or TokenUsage(from_cache=False)
        self.cached_token_usage = cached_token_usage or TokenUsage(from_cache=True)

    def __add__(self, other):
        if not isinstance(other, InterviewTokenUsage):
            raise ValueError(f"Can't add {type(other)} to InterviewTokenSummary")
        return InterviewTokenUsage(
            new_token_usage=self.new_token_usage + other.new_token_usage,
            cached_token_usage=self.cached_token_usage + other.cached_token_usage,
        )

    def __repr__(self):
        return f"InterviewTokenUsage(new_token_usage={self.new_token_usage}, cached_token_usage={self.cached_token_usage})"

    def cost(self, prices: TokenPricing):
        return self.new_token_usage.cost(prices)

    def saved(self, prices: TokenPricing):
        return self.cached_token_usage.cost(prices)



import textwrap
from typing import Dict
import json

import pandas as pd

from edsl.conjure.SurveyBuilder import SurveyBuilder
from edsl.conjure.utilities import RCodeSnippet


class SurveyBuilderSPSS(SurveyBuilder):
    header_r_code = RCodeSnippet(
        textwrap.dedent(
            """ 
    library(haven)
    library(jsonlite)
    args <- commandArgs(trailingOnly = TRUE)
    sav_file_path <- args[1]
    data <- read_sav(sav_file_path)
    """
        )
    )

    get_responses_r_code = header_r_code + RCodeSnippet(
        """
    write.csv(data, file = stdout(), row.names = FALSE)
    """
    )

    get_question_code_to_question_text_r_code = header_r_code + RCodeSnippet(
        textwrap.dedent(
            """
    question_codes <- colnames(data)
    question_labels <- as.character(sapply(data, function(x) {
        lbl <- attr(x, "label")
        if (is.null(lbl)) "" else lbl
    }))
    df <- data.frame(question_codes, question_labels)
    write.csv(df, file = stdout(), row.names = FALSE)
    """
        )
    )

    get_answer_code_to_answer_text_r_code = header_r_code + RCodeSnippet(
        textwrap.dedent(
            """
    convert_label <- function(d){
        df <- data.frame(name = names(d), value = as.numeric(d))
        json_representation <- toJSON(df, pretty = TRUE)
        json_representation
    }

    question_codes <- colnames(data)
    answer_codes <- sapply(data, function(x) convert_label(attr(x, "labels")))

    df <- data.frame(question_codes, answer_codes)
    write.csv(df, file = stdout(), row.names = FALSE)
    """
        )
    )

    def get_responses(self):
        """Returns a dataframe of responses.
        The structure should be a dictionary, where the keys are the question codes,
        and the values are the responses.

        For example, {"Q1": [1, 2, 3], "Q2": [4, 5, 6]}
        """
        df = self.get_responses_r_code(self.datafile_name)
        df.fillna("", inplace=True)
        df = df.astype(str)
        data_dict = df.to_dict(orient="list")
        return {k.lower(): v for k, v in data_dict.items()}

    def get_question_name_to_text(self) -> Dict:
        df = self.get_question_code_to_question_text_r_code(self.datafile_name)
        df.fillna("", inplace=True)
        df = df.astype(str)
        data_dict = df.to_dict(orient="list")

        question_codes = [q.lower() for q in data_dict["question_codes"]]
        question_text = data_dict["question_labels"]
        d = dict(zip(question_codes, question_text))
        try:
            assert len(d) == len(question_codes)
        except AssertionError:
            raise ValueError("Duplicate question codes found.")

        return d

    def get_question_name_to_answer_book(self):
        """Returns a dictionary mapping question codes to a dictionary mapping answer codes to answer text.

        e.g., {'q1': {1: 'yes', 2:'no'}}
        """
        df = self.get_answer_code_to_answer_text_r_code(self.datafile_name)
        df.fillna("", inplace=True)
        df = df.astype(str)
        data_dict = df.to_dict(orient="list")
        question_codes = [q.lower() for q in data_dict["question_codes"]]
        answer_dicts = []
        for answer_code in data_dict["answer_codes"]:
            try:
                parsed_list = json.loads(answer_code)
                value = {entry["value"]: entry["name"] for entry in parsed_list}
            except json.JSONDecodeError as e:
                value = answer_code
                print(
                    f"Warning: Could not parse answer_codes for {answer_code} as JSON. Using raw value instead."
                )
            answer_dicts.append(value)

        d = dict(zip(question_codes, answer_dicts))
        return d


if __name__ == "__main__":
    spss_builder = SurveyBuilderSPSS("job_satisfaction.sav", 100)



from __future__ import annotations
import random
import textwrap
from jinja2 import Template
from typing import Any, Optional, Union
from edsl.questions.QuestionBase import QuestionBase
from edsl.exceptions import QuestionAnswerValidationError
from edsl.scenarios import Scenario
from edsl.utilities.utilities import random_string
from edsl.questions.descriptors import (
    QuestionOptionsDescriptor,
    NumSelectionsDescriptor,
)


class QuestionRank(QuestionBase):
    """This question prompts the agent to rank options from a list."""

    question_type = "rank"
    question_options: list[str] = QuestionOptionsDescriptor()
    num_selections = NumSelectionsDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        num_selections: Optional[int] = None,
    ):
        """Initialize the question.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options the respondent should select from.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionRank.default_instructions`.
        :param min_selections: The minimum number of options that must be selected.
        :param max_selections: The maximum number of options that must be selected.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options
        self.num_selections = num_selections or len(question_options)

    ################
    # Answer methods
    ################
    def _validate_answer(self, answer: Any) -> dict[str, list[int]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value(answer, "answer", list)
        self._validate_answer_rank(answer)
        return answer

    def _translate_answer_code_to_answer(
        self, answer_codes, scenario: Scenario = None
    ) -> list[str]:
        """Translate the answer code to the actual answer."""
        scenario = scenario or Scenario()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            translated_codes.append(translated_options[int(answer_code)])
        return translated_codes

    def _simulate_answer(self, human_readable=True) -> dict[str, Union[int, str]]:
        """Simulate a valid answer for debugging purposes."""
        if human_readable:
            selected = random.sample(self.question_options, self.num_selections)
        else:
            selected = random.sample(
                range(len(self.question_options)), self.num_selections
            )
        answer = {
            "answer": selected,
            "comment": random_string(),
        }
        return answer

    @property
    def question_html_content(self) -> str:
        from jinja2 import Template

        question_html_content = Template(
            """
        <form id="rankForm">
            <p>{{ question_text }}</p>
            {% for option in question_options %}
            <div>
                <label for="{{ option }}">{{ option }}</label>
                <input type="number" id="{{ option }}" name="{{ question_name }}[{{ option }}]" value="0" min="1" max="{{ question_options|length }}" oninput="updateRankings()">
            </div>
            {% endfor %}
        </form>
        <script>
        function updateRankings() {
            let options = {{ question_options|length }};
            let values = [];
            let isValid = true;

            {% for option in question_options %}
            let value = parseInt(document.getElementById("{{ option }}").value) || 0;
            if (value > 0 && value <= options && !values.includes(value)) {
                values.push(value);
            } else if (value !== 0) {
                isValid = false;
            }
            {% endfor %}

            if (!isValid || values.length !== new Set(values).size) {
                document.getElementById("error").innerText = "Please enter unique and valid ranks for each option.";
            } else {
                document.getElementById("error").innerText = "";
            }
        }
        </script>
        <p id="error" style="color: red;"></p>
        """
        ).render(
            question_name=self.question_name,
            question_text=self.question_text,
            question_options=self.question_options,
        )
        return question_html_content

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionRank:
        """Return an example question."""
        return cls(
            question_name="rank_foods",
            question_text="Rank your favorite foods.",
            question_options=["Pizza", "Pasta", "Salad", "Soup"],
            num_selections=2,
        )


def main():
    """Show example usage."""
    from edsl.questions.QuestionRank import QuestionRank

    q = QuestionRank.example()
    q.question_text
    q.question_name
    q.question_options
    q.num_selections
    # validate an answer
    answer = {"answer": [0, 1], "comment": "I like pizza and pasta."}
    q._validate_answer(answer)
    # translate an answer code to an answer
    q._translate_answer_code_to_answer([0, 1])
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""
The `Cache` class is used to store responses from a language model.
"""

from __future__ import annotations
import json
import os
import warnings
from typing import Optional, Union

from edsl.config import CONFIG
from edsl.data.CacheEntry import CacheEntry
from edsl.data.SQLiteDict import SQLiteDict
from edsl.Base import Base

from edsl.utilities.decorators import (
    add_edsl_version,
    remove_edsl_version,
)


class Cache(Base):
    """
    A class that represents a cache of responses from a language model.

    :param data: The data to initialize the cache with.
    :param remote: Whether to sync the Cache with the server.
    :param immediate_write: Whether to write to the cache immediately after storing a new entry.

    Deprecated:

    :param method: The method of storage to use for the cache.
    """

    data = {}

    def __init__(
        self,
        *,
        data: Optional[Union[SQLiteDict, dict]] = None,
        remote: bool = False,
        immediate_write: bool = True,
        method=None,
    ):
        """
        Create two dictionaries to store the cache data.

        """
        self.data = data or {}
        # self.data_at_init = data or {}
        self.fetched_data = {}
        self.remote = remote
        self.immediate_write = immediate_write
        self.method = method
        self.new_entries = {}
        self.new_entries_to_write_later = {}
        self.coop = None
        self._perform_checks()

    def rich_print(sefl):
        pass
        # raise NotImplementedError("This method is not implemented yet.")

    def code(sefl):
        pass
        # raise NotImplementedError("This method is not implemented yet.")

    def keys(self):
        return list(self.data.keys())

    def values(self):
        return list(self.data.values())

    def new_entries_cache(self) -> Cache:
        """Return a new Cache object with the new entries."""
        return Cache(data={**self.new_entries, **self.fetched_data})

    def _perform_checks(self):
        """Perform checks on the cache."""
        if any(not isinstance(value, CacheEntry) for value in self.data.values()):
            raise Exception("Not all values are CacheEntry instances")
        if self.method is not None:
            warnings.warn("Argument `method` is deprecated", DeprecationWarning)
        if self.remote:
            from edsl.coop import Coop

            self.coop = Coop()

    ####################
    # READ/WRITE
    ####################
    def fetch(
        self,
        *,
        model: str,
        parameters: dict,
        system_prompt: str,
        user_prompt: str,
        iteration: int,
    ) -> Union[None, str]:
        """
        Fetch a value (LLM output) from the cache.

        :param model: The name of the language model.
        :param parameters: The model parameters.
        :param system_prompt: The system prompt.
        :param user_prompt: The user prompt.
        :param iteration: The iteration number.

        Return None if the response is not found.

        >>> c = Cache()
        >>> c.fetch(model="gpt-3", parameters="default", system_prompt="Hello", user_prompt="Hi", iteration=1) is None
        True


        """
        key = CacheEntry.gen_key(
            model=model,
            parameters=parameters,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            iteration=iteration,
        )
        entry = self.data.get(key, None)
        if entry is not None:
            self.fetched_data[key] = entry
        return None if entry is None else entry.output

    def store(
        self,
        model: str,
        parameters: str,
        system_prompt: str,
        user_prompt: str,
        response: str,
        iteration: int,
    ) -> str:
        """
        Add a new key-value pair to the cache.

        * Key is a hash of the input parameters.
        * Output is the response from the language model.

        How it works:

        * The key-value pair is added to `self.new_entries`
        * If `immediate_write` is True , the key-value pair is added to `self.data`
        * If `immediate_write` is False, the key-value pair is added to `self.new_entries_to_write_later`
        """
        entry = CacheEntry(
            model=model,
            parameters=parameters,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            output=json.dumps(response),
            iteration=iteration,
        )
        key = entry.key
        self.new_entries[key] = entry
        if self.immediate_write:
            self.data[key] = entry
        else:
            self.new_entries_to_write_later[key] = entry
        return key

    def add_from_dict(
        self, new_data: dict[str, CacheEntry], write_now: Optional[bool] = True
    ) -> None:
        """
        Add entries to the cache from a dictionary.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        for key, value in new_data.items():
            if key in self.data:
                if value != self.data[key]:
                    raise Exception("Mismatch in values")
            if not isinstance(value, CacheEntry):
                raise Exception(f"Wrong type - the observed type is {type(value)}")

        self.new_entries.update(new_data)
        if write_now:
            self.data.update(new_data)
        else:
            self.new_entries_to_write_later.update(new_data)

    def add_from_jsonl(self, filename: str, write_now: Optional[bool] = True) -> None:
        """
        Add entries to the cache from a JSONL.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        with open(filename, "a+") as f:
            f.seek(0)
            lines = f.readlines()
        new_data = {}
        for line in lines:
            d = json.loads(line)
            key = list(d.keys())[0]
            value = list(d.values())[0]
            new_data[key] = CacheEntry(**value)
        self.add_from_dict(new_data=new_data, write_now=write_now)

    def add_from_sqlite(self, db_path: str, write_now: Optional[bool] = True):
        """
        Add entries to the cache from an SQLite database.

        :param write_now: Whether to write to the cache immediately (similar to `immediate_write`).
        """
        db = SQLiteDict(db_path)
        new_data = {}
        for key, value in db.items():
            new_data[key] = CacheEntry(**value)
        self.add_from_dict(new_data=new_data, write_now=write_now)

    @classmethod
    def from_sqlite_db(cls, db_path: str) -> Cache:
        """
        Construct a Cache from a SQLite database.
        """
        return cls(data=SQLiteDict(db_path))

    @classmethod
    def from_local_cache(cls) -> Cache:
        """
        Construct a Cache from a local cache file.
        """
        from edsl.config import CONFIG

        CACHE_PATH = CONFIG.get("EDSL_DATABASE_PATH")
        path = CACHE_PATH.replace("sqlite:///", "")
        db_path = os.path.join(os.path.dirname(path), "data.db")
        return cls.from_sqlite_db(db_path=db_path)

    @classmethod
    def from_jsonl(cls, jsonlfile: str, db_path: Optional[str] = None) -> Cache:
        """
        Construct a Cache from a JSONL file.

        :param jsonlfile: The path to the JSONL file of cache entries.
        :param db_path: The path to the SQLite database used to store the cache.

        * If `db_path` is None, the cache will be stored in memory, as a dictionary.
        * If `db_path` is provided, the cache will be stored in an SQLite database.
        """
        # if a file doesn't exist at jsonfile, throw an error
        if not os.path.exists(jsonlfile):
            raise FileNotFoundError(f"File {jsonlfile} not found")

        if db_path is None:
            data = {}
        else:
            data = SQLiteDict(db_path)

        cache = Cache(data=data)
        cache.add_from_jsonl(jsonlfile)
        return cache

    def write_sqlite_db(self, db_path: str) -> None:
        """
        Write the cache to an SQLite database.
        """
        ## TODO: Check to make sure not over-writing (?)
        ## Should be added to SQLiteDict constructor (?)
        new_data = SQLiteDict(db_path)
        for key, value in self.data.items():
            new_data[key] = value

    def write_jsonl(self, filename: str) -> None:
        """
        Write the cache to a JSONL file.
        """
        path = os.path.join(os.getcwd(), filename)
        with open(path, "w") as f:
            for key, value in self.data.items():
                f.write(json.dumps({key: value.to_dict()}) + "\n")

    ####################
    # REMOTE
    ####################
    # TODO: Make this work
    # - Need to decide whether the cache belongs to a user and what can be shared
    # - I.e., some cache entries? all or nothing?
    @classmethod
    def from_url(cls, db_path=None) -> Cache:
        """
        Construct a Cache object from a remote.
        """
        # ...do something here
        # return Cache(data=db)
        pass

    def __enter__(self):
        """
        Run when a context is entered.
        """
        if self.remote:
            print("Syncing local and remote caches")
            exclude_keys = list(self.data.keys())
            cache_entries = self.coop.get_cache_entries(exclude_keys)
            self.add_from_dict({c.key: c for c in cache_entries}, write_now=True)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Run when a context is exited.
        """
        for key, entry in self.new_entries_to_write_later.items():
            self.data[key] = entry
        if self.remote:
            _ = self.coop.create_cache_entries(cache_dict=self.new_entries)

    ####################
    # DUNDER / USEFUL
    ####################
    @add_edsl_version
    def to_dict(self) -> dict:
        """Return the Cache as a dictionary."""
        return {k: v.to_dict() for k, v in self.data.items()}

    def _repr_html_(self):
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict())

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data) -> Cache:
        """Construct a Cache from a dictionary."""
        newdata = {k: CacheEntry.from_dict(v) for k, v in data.items()}
        return cls(data=newdata)

    def __len__(self):
        """Return the number of CacheEntry objects in the Cache."""
        return len(self.data)

    # TODO: Same inputs could give different results and this could be useful
    # can't distinguish unless we do the ε trick or vary iterations
    def __eq__(self, other_cache: "Cache") -> bool:
        """
        Check if two Cache objects are equal.
        Does not verify their values are equal, only that they have the same keys.
        """
        if not isinstance(other_cache, Cache):
            return False
        return set(self.data.keys()) == set(other_cache.data.keys())

    def __add__(self, other: "Cache"):
        """
        Combine two caches.
        """
        if not isinstance(other, Cache):
            raise ValueError("Can only add two caches together")
        self.data.update(other.data)
        return self

    def __repr__(self):
        """
        Return a string representation of the Cache object.
        """
        return f"Cache(data = {repr(self.data)}, immediate_write={self.immediate_write}, remote={self.remote})"

    ####################
    # EXAMPLES
    ####################
    def fetch_input_example(self) -> dict:
        """
        Create an example input for a 'fetch' operation.
        """
        return CacheEntry.fetch_input_example()

    def to_html(self):
        # json_str = json.dumps(self.data, indent=4)
        d = {k: v.to_dict() for k, v in self.data.items()}
        for key, value in d.items():
            for k, v in value.items():
                if isinstance(v, dict):
                    d[key][k] = {kk: str(vv) for kk, vv in v.items()}
                else:
                    d[key][k] = str(v)

        json_str = json.dumps(d, indent=4)

        # HTML template with the JSON string embedded
        html = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Display JSON</title>
        </head>
        <body>
            <pre id="jsonData"></pre>
            <script>
                var json = {json_str};

                // JSON.stringify with spacing to format
                document.getElementById('jsonData').textContent = JSON.stringify(json, null, 4);
            </script>
        </body>
        </html>
        """
        return html

    def view(self) -> None:
        """View the Cache in a new browser tab."""
        import tempfile
        import webbrowser

        html_content = self.to_html()
        # Create a temporary file to hold the HTML
        with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html") as tmpfile:
            tmpfile.write(html_content)
            # Get the path to the temporary file
            filepath = tmpfile.name

        # Open the HTML file in a new browser tab
        webbrowser.open("file://" + filepath)

    @classmethod
    def example(cls) -> Cache:
        """
        Return an example Cache.
        The example Cache has one entry.
        """
        return cls(data={CacheEntry.example().key: CacheEntry.example()})


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""Class for creating prompts to be used in a survey."""

from __future__ import annotations
from typing import Optional
from abc import ABC
from typing import Any, List

from rich.table import Table
from jinja2 import Template, Environment, meta, TemplateSyntaxError

from edsl.exceptions.prompts import TemplateRenderError
from edsl.prompts.prompt_config import (
    C2A,
    names_to_component_types,
    ComponentTypes,
    NEGATIVE_INFINITY,
)
from edsl.prompts.registry import RegisterPromptsMeta
from edsl.Base import PersistenceMixin, RichPrintingMixin

MAX_NESTING = 100


class PromptBase(
    PersistenceMixin, RichPrintingMixin, ABC, metaclass=RegisterPromptsMeta
):
    """Class for creating a prompt to be used in a survey."""

    default_instructions: Optional[str] = "Do good things, friendly LLM!"
    component_type = ComponentTypes.GENERIC

    def _repr_html_(self):
        """Return an HTML representation of the Prompt."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.to_dict())

    @classmethod
    def prompt_attributes(cls) -> List[str]:
        """Return the prompt class attributes."""
        return {k: v for k, v in cls.__dict__.items() if not k.startswith("_")}

    def __init__(self, text: Optional[str] = None):
        """Create a `Prompt` object.

        :param text: The text of the prompt.
        """
        if text is None:
            if hasattr(self, "default_instructions"):
                text = self.default_instructions
            else:
                text = ""
        self._text = text

    @classmethod
    def from_txt(cls, filename: str) -> PromptBase:
        """Create a `Prompt` from text.

        :param text: The text of the prompt.
        """
        with open(filename, "r") as f:
            text = f.read()
        return cls(text=text)

    @property
    def text(self):
        """Return the `Prompt` text."""
        return self._text

    def __add__(self, other_prompt):
        """Add two prompts together.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt("How are you?")
        >>> p + p2
        Prompt(text='Hello, {{person}}How are you?')

        >>> p + "How are you?"
        Prompt(text='Hello, {{person}}How are you?')
        """
        if isinstance(other_prompt, str):
            return self.__class__(self.text + other_prompt)
        else:
            return self.__class__(text=self.text + other_prompt.text)

    def __str__(self):
        """Return the `Prompt` text.

        Example:
        >>> p = Prompt("Hello, {{person}}")
        >>> str(p)
        'Hello, {{person}}'
        """
        return self.text

    def __contains__(self, text_to_check):
        """Check if the text_to_check is in the `Prompt` text.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> "person" in p
        True
        >>> "person2" in p
        False
        """
        return text_to_check in self.text

    def __repr__(self):
        """Return the `Prompt` text.

        Example:
        >>> p = Prompt("Hello, {{person}}")
        >>> p
        Prompt(text='Hello, {{person}}')
        """
        return f"Prompt(text='{self.text}')"

    def template_variables(self) -> list[str]:
        """Return the the variables in the template.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.template_variables()
        ['person']

        """
        return self._template_variables(self.text)

    @staticmethod
    def _template_variables(template: str) -> list[str]:
        """Find and return the template variables.

        :param template: The template to find the variables in.

        """
        env = Environment()
        ast = env.parse(template)
        return list(meta.find_undeclared_variables(ast))

    def undefined_template_variables(self, replacement_dict: dict):
        """Return the variables in the template that are not in the replacement_dict.

        :param replacement_dict: A dictionary of replacements to populate the template.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.undefined_template_variables({"person": "John"})
        []

        >>> p = Prompt("Hello, {{title}} {{person}}")
        >>> p.undefined_template_variables({"person": "John"})
        ['title']
        """
        return [var for var in self.template_variables() if var not in replacement_dict]

    def unused_traits(self, traits: dict):
        """Return the traits that are not used in the template."""
        return [trait for trait in traits if trait not in self.template_variables()]

    @property
    def has_variables(self) -> bool:
        """Return True if the prompt has variables.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.has_variables
        True

        >>> p = Prompt("Hello, person")
        >>> p.has_variables
        False
        """
        return len(self.template_variables()) > 0

    def render(self, primary_replacement: dict, **additional_replacements) -> str:
        """Render the prompt with the replacements.

        :param primary_replacement: The primary replacement dictionary.
        :param additional_replacements: Additional replacement dictionaries.

        >>> p = Prompt("Hello, {{person}}")
        >>> p.render({"person": "John"})
        'Hello, John'

        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Horton"})
        'Hello, Mr. Horton'

        >>> p.render({"person": "Mr. {{last_name}}", "last_name": "Ho{{letter}}ton"}, max_nesting = 1)
        'Hello, Mr. Horton'
        """
        new_text = self._render(
            self.text, primary_replacement, **additional_replacements
        )
        return self.__class__(text=new_text)

    @staticmethod
    def _render(
        text: str, primary_replacement, **additional_replacements
    ) -> "PromptBase":
        """Render the template text with variables replaced from the provided named dictionaries.

        :param text: The text to render.
        :param primary_replacement: The primary replacement dictionary.
        :param additional_replacements: Additional replacement dictionaries.

        Allows for nested variable resolution up to a specified maximum nesting depth.

        Example:

        >>> codebook = {"age": "Age"}
        >>> p = Prompt("You are an agent named {{ name }}. {{ codebook['age']}}: {{ age }}")
        >>> p.render({"name": "John", "age": 44}, codebook=codebook)
        'You are an agent named John. Age: 44'
        """
        try:
            previous_text = None
            for _ in range(MAX_NESTING):
                rendered_text = Template(text).render(
                    primary_replacement, **additional_replacements
                )
                if rendered_text == previous_text:
                    # No more changes, so return the rendered text
                    return rendered_text
                previous_text = text
                text = rendered_text

            # If the loop exits without returning, it indicates too much nesting
            raise TemplateRenderError(
                "Too much nesting - you created an infinite loop here, pal"
            )
        except TemplateSyntaxError as e:
            raise TemplateRenderError(f"Template syntax error: {e}")

    def to_dict(self) -> dict[str, Any]:
        """Return the `Prompt` as a dictionary.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p.to_dict()
        {'text': 'Hello, {{person}}', 'class_name': 'Prompt'}

        """
        return {"text": self.text, "class_name": self.__class__.__name__}

    @classmethod
    def from_dict(cls, data) -> PromptBase:
        """Create a `Prompt` from a dictionary.

        Example:

        >>> p = Prompt("Hello, {{person}}")
        >>> p2 = Prompt.from_dict(p.to_dict())
        >>> p2
        Prompt(text='Hello, {{person}}')

        """
        class_name = data["class_name"]
        cls = RegisterPromptsMeta._registry.get(class_name, Prompt)
        return cls(text=data["text"])

    def rich_print(self):
        """Display an object as a table."""
        table = Table(title="Prompt")
        table.add_column("Attribute", style="bold")
        table.add_column("Value")

        to_display = self.__dict__.copy()
        for attr_name, attr_value in to_display.items():
            table.add_row(attr_name, repr(attr_value))
        table.add_row("Component type", str(self.component_type))
        table.add_row("Model", str(getattr(self, "model", "Not specified")))
        return table

    @classmethod
    def example(cls):
        """Return an example of the prompt."""
        return cls(cls.default_instructions)


class Prompt(PromptBase):
    """A prompt to be used in a survey."""

    component_type = ComponentTypes.GENERIC


from edsl.prompts.library.question_multiple_choice import *
from edsl.prompts.library.agent_instructions import *
from edsl.prompts.library.agent_persona import *

from edsl.prompts.library.question_budget import *
from edsl.prompts.library.question_checkbox import *
from edsl.prompts.library.question_freetext import *
from edsl.prompts.library.question_linear_scale import *
from edsl.prompts.library.question_numerical import *
from edsl.prompts.library.question_rank import *
from edsl.prompts.library.question_extract import *
from edsl.prompts.library.question_list import *


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Dict

import pandas as pd

from edsl.conjure.SurveyBuilder import SurveyBuilder
from edsl.conjure.utilities import RCodeSnippet


class SurveyBuilderCSV(SurveyBuilder):
    @staticmethod
    def get_dataframe(datafile_name):
        return pd.read_csv(datafile_name)

    def get_responses(self) -> Dict:
        """Returns a dataframe of responses by reading the datafile_name.

        The structure should be a dictionary, where the keys are the question codes,
        and the values are the responses.

        For example, {"Q1": [1, 2, 3], "Q2": [4, 5, 6]}

        >>> sb = SurveyBuilderCSV.example()
        >>> sb.get_responses()
        {'q1': ['1', '4'], 'q2': ['2', '5'], 'q3': ['3', '6']}

        """
        df = self.get_dataframe(self.datafile_name)
        df.fillna("", inplace=True)
        df = df.astype(str)
        data_dict = df.to_dict(orient="list")
        return {k.lower(): v for k, v in data_dict.items()}

    def get_question_name_to_text(self) -> Dict:
        """
        Get the question name to text mapping.

        >>> sb = SurveyBuilderCSV.example()
        >>> sb.get_question_name_to_text()
        {'Q1': 'Q1', 'Q2': 'Q2', 'Q3': 'Q3'}

        """
        d = {}
        df = self.get_dataframe(self.datafile_name)
        for col in df.columns:
            if col in self.lookup_dict():
                d[col] = self.lookup_dict()[col]
            else:
                d[col] = col

        return d

    def get_question_name_to_answer_book(self):
        """Returns a dictionary mapping question codes to a dictionary mapping answer codes to answer text."""
        d = self.get_question_name_to_text()
        return {k: {} for k, v in d.items()}

    @classmethod
    def example(cls):
        import tempfile

        named_temp_file = tempfile.NamedTemporaryFile(delete=False)
        named_temp_file.write(b"Q1,Q2,Q3\n1,2,3\n4,5,6\n")
        named_temp_file.close()
        return cls(named_temp_file.name)


class SurveyBuilderStata(SurveyBuilderCSV):
    @staticmethod
    def get_dataframe(datafile_name):
        return pd.read_stata(datafile_name)


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # sb = SurveyBuilderCSV("responses.csv")
    # sb.save("podcast_survey")



"""Mixin class for ggplot2 plotting."""

import subprocess
import pandas as pd
import tempfile
from typing import Optional
import matplotlib.pyplot as plt
import matplotlib.image as mpimg


class ResultsGGMixin:
    """Mixin class for ggplot2 plotting."""

    def ggplot2(
        self,
        ggplot_code: str,
        filename: str = None,
        shape="wide",
        sql: str = None,
        remove_prefix: bool = True,
        debug: bool = False,
        height=4,
        width=6,
        format="svg",
        factor_orders: Optional[dict] = None,
    ):
        """Create a ggplot2 plot from a DataFrame.

        :param ggplot_code: The ggplot2 code to execute.
        :param filename: The filename to save the plot to.
        :param shape: The shape of the data in the DataFrame (wide or long).
        :param sql: The SQL query to execute beforehand to manipulate the data.
        :param remove_prefix: Whether to remove the prefix from the column names.
        :param debug: Whether to print the R code instead of executing it.
        :param height: The height of the plot in inches.
        :param width: The width of the plot in inches.
        :param format: The format to save the plot in (png or svg).
        :param factor_orders: A dictionary of factor columns and their order.
        """

        if sql == None:
            sql = "select * from self"

        if shape == "long":
            df = self.sql(sql, shape="long")
        elif shape == "wide":
            df = self.sql(sql, shape="wide", remove_prefix=remove_prefix)

        # Convert DataFrame to CSV format
        csv_data = df.to_csv(index=False)

        # Embed the CSV data within the R script
        csv_data_escaped = csv_data.replace("\n", "\\n").replace("'", "\\'")
        read_csv_code = f"self <- read.csv(text = '{csv_data_escaped}', sep = ',')\n"

        if factor_orders is not None:
            for factor, order in factor_orders.items():
                # read_csv_code += f"""self${{{factor}}} <- factor(self${{{factor}}}, levels=c({','.join(['"{}"'.format(x) for x in order])}))"""

                level_string = ", ".join([f'"{x}"' for x in order])
                read_csv_code += (
                    f"self${factor} <- factor(self${factor}, levels=c({level_string}))"
                )
                read_csv_code += "\n"

        # Load ggplot2 library
        load_ggplot2 = "library(ggplot2)\n"

        # Check if a filename is provided for the plot, if not create a temporary one
        if not filename:
            filename = tempfile.mktemp(suffix=f".{format}")

        # Combine all R script parts
        full_r_code = load_ggplot2 + read_csv_code + ggplot_code

        # Add command to save the plot to a file
        full_r_code += f'\nggsave("{filename}", plot = last_plot(), width = {width}, height = {height}, device = "{format}")'

        if debug:
            print(full_r_code)
            return

        result = subprocess.run(
            ["Rscript", "-"],
            input=full_r_code,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if result.returncode != 0:
            if result.returncode == 127:  # 'command not found'
                raise RuntimeError(
                    "Rscript is probably not installed. Please install R from https://cran.r-project.org/"
                )
            else:
                raise RuntimeError(
                    f"An error occurred while running Rscript: {result.stderr}"
                )

        if result.stderr:
            print("Error in R script:", result.stderr)
        else:
            self._display_plot(filename, width, height)

    def _display_plot(self, filename: str, width: float, height: float):
        """Display the plot in the notebook."""
        if filename.endswith(".png"):
            img = mpimg.imread(filename)
            plt.figure(
                figsize=(width, height)
            )  # Set the figure size (width, height) in inches
            plt.imshow(img)
            plt.axis("off")
            plt.show()
        elif filename.endswith(".svg"):
            from IPython.display import SVG, display

            display(SVG(filename=filename))
        else:
            print("Unsupported file format. Please provide a PNG or SVG file.")



import aiohttp
import json
import requests
from typing import Any
from edsl.inference_services.InferenceServiceABC import InferenceServiceABC
from edsl.language_models import LanguageModel


class DeepInfraService(InferenceServiceABC):
    """DeepInfra service class."""

    _inference_service_ = "deep_infra"
    _env_key_name_ = "DEEP_INFRA_API_KEY"

    @classmethod
    def available(cls):
        text_models = cls.full_details_available()
        return [m["model_name"] for m in text_models]

    @classmethod
    def full_details_available(cls, verbose=False):
        url = "https://api.deepinfra.com/models/list"
        response = requests.get(url)
        if response.status_code == 200:
            text_generation_models = [
                r for r in response.json() if r["type"] == "text-generation"
            ]
            from rich import print_json
            import json

            if verbose:
                print_json(json.dumps(text_generation_models))
            return text_generation_models
        else:
            return f"Failed to fetch data: Status code {response.status_code}"

    @classmethod
    def create_model(cls, model_name: str, model_class_name=None) -> LanguageModel:
        base_url = "https://api.deepinfra.com/v1/inference/"
        if model_class_name is None:
            model_class_name = cls.to_class_name(model_name)
        url = f"{base_url}{model_name}"

        class LLM(LanguageModel):
            _inference_service_ = cls._inference_service_
            _model_ = model_name
            _parameters_ = {
                "temperature": 0.7,
                "top_p": 0.2,
                "top_k": 0.1,
                "max_new_tokens": 512,
                "stopSequences": [],
            }

            async def async_execute_model_call(
                self, user_prompt: str, system_prompt: str = ""
            ) -> dict[str, Any]:
                self.url = url
                headers = {
                    "Content-Type": "application/json",
                    "Authorization": f"bearer {self.api_token}",
                }
                # don't mess w/ the newlines
                data = {
                    "input": f"""
                    [INST]<<SYS>>
                    {system_prompt}
                    <<SYS>>{user_prompt}[/INST]
                    """,
                    "stream": False,
                    "temperature": self.temperature,
                    "top_p": self.top_p,
                    "top_k": self.top_k,
                    "max_new_tokens": self.max_new_tokens,
                }
                async with aiohttp.ClientSession() as session:
                    async with session.post(
                        self.url, headers=headers, data=json.dumps(data)
                    ) as response:
                        raw_response_text = await response.text()
                        return json.loads(raw_response_text)

            def parse_response(self, raw_response: dict[str, Any]) -> str:
                if "results" not in raw_response:
                    raise Exception(
                        f"Deep Infra response does not contain 'results' key: {raw_response}"
                    )
                if "generated_text" not in raw_response["results"][0]:
                    raise Exception(
                        f"Deep Infra response does not contain 'generate_text' key: {raw_response['results'][0]}"
                    )
                return raw_response["results"][0]["generated_text"]

        LLM.__name__ = model_class_name

        return LLM



from __future__ import annotations
import time
import asyncio
import textwrap
from contextlib import contextmanager

from typing import Coroutine, List, AsyncGenerator, Optional, Union

from rich.live import Live
from rich.console import Console

from edsl import shared_globals
from edsl.results import Results, Result

from edsl.jobs.interviews.Interview import Interview
from edsl.utilities.decorators import jupyter_nb_handler
from edsl.jobs.Jobs import Jobs
from edsl.jobs.runners.JobsRunnerStatusMixin import JobsRunnerStatusMixin
from edsl.language_models import LanguageModel
from edsl.data.Cache import Cache

from edsl.jobs.tasks.TaskHistory import TaskHistory
from edsl.jobs.buckets.BucketCollection import BucketCollection


class JobsRunnerAsyncio(JobsRunnerStatusMixin):
    """A class for running a collection of interviews asynchronously.

    It gets instaniated from a Jobs object.
    The Jobs object is a collection of interviews that are to be run.
    """

    def __init__(self, jobs: Jobs):
        self.jobs = jobs

        self.interviews: List["Interview"] = jobs.interviews()
        self.bucket_collection: "BucketCollection" = jobs.bucket_collection
        self.total_interviews: List["Interview"] = []

    async def run_async(
        self,
        cache: Cache,
        n: int = 1,
        debug: bool = False,
        stop_on_exception: bool = False,
        sidecar_model: "LanguageModel" = None,
        total_interviews: Optional[List["Interview"]] = None,
    ) -> AsyncGenerator[Result, None]:
        """Creates the tasks, runs them asynchronously, and returns the results as a Results object.

        Completed tasks are yielded as they are completed.

        :param n: how many times to run each interview
        :param debug:
        :param stop_on_exception:
        """
        tasks = []
        if total_interviews:
            self.total_interviews = total_interviews
        else:
            self._populate_total_interviews(
                n=n
            )  # Populate self.total_interviews before creating tasks

        for interview in self.total_interviews:
            interviewing_task = self._build_interview_task(
                interview=interview,
                debug=debug,
                stop_on_exception=stop_on_exception,
                sidecar_model=sidecar_model,
            )
            tasks.append(asyncio.create_task(interviewing_task))

        for task in asyncio.as_completed(tasks):
            result = await task
            yield result

    def _populate_total_interviews(self, n: int = 1) -> None:
        """Populates self.total_interviews with n copies of each interview.

        :param n: how many times to run each interview.
        """
        # TODO: Why not return a list of interviews instead of modifying the object?

        self.total_interviews = []
        for interview in self.interviews:
            for iteration in range(n):
                if iteration > 0:
                    new_interview = interview.duplicate(
                        iteration=iteration, cache=self.cache
                    )
                    self.total_interviews.append(new_interview)
                else:
                    interview.cache = (
                        self.cache
                    )  # set the cache for the first interview
                    self.total_interviews.append(interview)

    async def _build_interview_task(
        self,
        *,
        interview: Interview,
        debug: bool,
        stop_on_exception: bool = False,
        sidecar_model: Optional[LanguageModel] = None,
    ) -> Result:
        """Conducts an interview and returns the result.

        :param interview: the interview to conduct
        :param debug: prints debug messages
        :param stop_on_exception: stops the interview if an exception is raised
        :param sidecar_model: a language model to use in addition to the interview's model
        """
        # the model buckets are used to track usage rates
        model_buckets = self.bucket_collection[interview.model]

        # get the results of the interview
        answer, valid_results = await interview.async_conduct_interview(
            debug=debug,
            model_buckets=model_buckets,
            stop_on_exception=stop_on_exception,
            sidecar_model=sidecar_model,
        )

        # we should have a valid result for each question
        answer_key_names = {k for k in set(answer.keys()) if not k.endswith("_comment")}

        assert len(valid_results) == len(answer_key_names)

        # TODO: move this down into Interview
        question_name_to_prompts = dict({})
        for result in valid_results:
            question_name = result["question_name"]
            question_name_to_prompts[question_name] = {
                "user_prompt": result["prompts"]["user_prompt"],
                "system_prompt": result["prompts"]["system_prompt"],
            }

        prompt_dictionary = {}
        for answer_key_name in answer_key_names:
            prompt_dictionary[
                answer_key_name + "_user_prompt"
            ] = question_name_to_prompts[answer_key_name]["user_prompt"]
            prompt_dictionary[
                answer_key_name + "_system_prompt"
            ] = question_name_to_prompts[answer_key_name]["system_prompt"]

        raw_model_results_dictionary = {}
        for result in valid_results:
            question_name = result["question_name"]
            raw_model_results_dictionary[
                question_name + "_raw_model_response"
            ] = result["raw_model_response"]

        result = Result(
            agent=interview.agent,
            scenario=interview.scenario,
            model=interview.model,
            iteration=interview.iteration,
            answer=answer,
            prompt=prompt_dictionary,
            raw_model_response=raw_model_results_dictionary,
            survey=interview.survey,
        )
        return result

    @property
    def elapsed_time(self):
        return time.monotonic() - self.start_time

    @jupyter_nb_handler
    async def run(
        self,
        cache: Union[Cache, False, None],
        n: int = 1,
        debug: bool = False,
        stop_on_exception: bool = False,
        progress_bar: bool = False,
        sidecar_model: Optional[LanguageModel] = None,
        print_exceptions: bool = True,
    ) -> "Coroutine":
        """Runs a collection of interviews, handling both async and sync contexts."""
        console = Console()
        self.results = []
        self.start_time = time.monotonic()
        self.completed = False
        self.cache = cache
        self.sidecar_model = sidecar_model

        def generate_table():
            return self.status_table(self.results, self.elapsed_time)

        @contextmanager
        def no_op_cm():
            """A no-op context manager with a dummy update method."""
            yield DummyLive()

        class DummyLive:
            def update(self, *args, **kwargs):
                """A dummy update method that does nothing."""
                pass

        progress_bar_context = (
            Live(generate_table(), console=console, refresh_per_second=5)
            if progress_bar
            else no_op_cm()
        )

        with cache as c:
            with progress_bar_context as live:

                async def update_progress_bar():
                    """Updates the progress bar at fixed intervals."""
                    while True:
                        live.update(generate_table())
                        await asyncio.sleep(0.1)  # Update interval
                        if self.completed:
                            break

                async def process_results():
                    """Processes results from interviews."""
                    async for result in self.run_async(
                        n=n,
                        debug=debug,
                        stop_on_exception=stop_on_exception,
                        cache=c,
                        sidecar_model=sidecar_model,
                    ):
                        self.results.append(result)
                        live.update(generate_table())
                    self.completed = True

                progress_task = asyncio.create_task(update_progress_bar())

                try:
                    await asyncio.gather(process_results(), progress_task)
                except asyncio.CancelledError:
                    pass
                finally:
                    progress_task.cancel()  # Cancel the progress_task when process_results is done
                    await progress_task

                    await asyncio.sleep(1)  # short delay to show the final status

                    # one more update
                    live.update(generate_table())

        results = Results(survey=self.jobs.survey, data=self.results)
        task_history = TaskHistory(self.total_interviews, include_traceback=False)
        results.task_history = task_history

        results.has_exceptions = task_history.has_exceptions

        if results.has_exceptions:
            failed_interviews = [
                interview.duplicate(
                    iteration=interview.iteration, cache=interview.cache
                )
                for interview in self.total_interviews
                if interview.has_exceptions
            ]
            results.failed_jobs = Jobs.from_interviews(
                [interview for interview in failed_interviews]
            )
            if print_exceptions:
                msg = f"Exceptions were raised in {len(results.task_history.indices)} out of {len(self.total_interviews)} interviews.\n"

                if len(results.task_history.indices) > 5:
                    msg += f"Exceptions were raised in the following interviews: {results.task_history.indices}.\n"

                shared_globals["edsl_runner_exceptions"] = task_history
                print(msg)
                task_history.html(cta="Open report to see details.")
                print(
                    "Also see: https://docs.expectedparrot.com/en/latest/exceptions.html"
                )

        return results



from edsl.conjure.SurveyBuilder import SurveyBuilder
from edsl.conjure.SurveyBuilderCSV import SurveyBuilderCSV
from edsl.conjure.SurveyBuilderCSV import SurveyBuilderStata
from edsl.conjure.SurveyBuilderSPSS import SurveyBuilderSPSS



import subprocess
from io import StringIO

import pandas as pd


class Missing:
    def __repr__(self):
        return "Missing()"

    def __str__(self):
        return "Missing()"

    def value(self):
        return "missing"


def convert_value(x):
    try:
        float_val = float(x)
        if float_val.is_integer():
            return int(float_val)
        else:
            return float_val
    except ValueError:
        if len(x) == 0:
            return Missing().value()
        else:
            return str(x)


class RCodeSnippet:
    def __init__(self, r_code):
        self.r_code = r_code

    def __call__(self, data_file_name):
        return self.run_R_stdin(self.r_code, data_file_name)

    def __add__(self, other):
        return RCodeSnippet(self.r_code + other.r_code)

    def write_to_file(self, filename) -> None:
        """Writes the R code to a file; useful for debugging."""
        if filename.endswith(".R") or filename.endswith(".r"):
            pass
        else:
            filename += ".R"

        with open(filename, "w") as f:
            f.write(self.r_code)

    @staticmethod
    def run_R_stdin(r_code, data_file_name, transform_func=lambda x: pd.read_csv(x)):
        """Runs an R script and returns the stdout as a string."""
        cmd = ["Rscript", "-e", r_code, data_file_name]
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        stdout, stderr = process.communicate()
        if stderr != "":
            print("Warning: stderr is not empty.")
            print(f"Problem running: {r_code}")
            raise Exception(stderr)
        return transform_func(StringIO(stdout))



# """The Jobs class is a collection of agents, scenarios and models and one survey."""
from __future__ import annotations
import os
from itertools import product
from typing import Optional, Union, Sequence, Generator
from edsl import Model
from edsl.agents import Agent
from edsl.Base import Base
from edsl.data.Cache import Cache
from edsl.data.CacheHandler import CacheHandler
from edsl.results.Dataset import Dataset

from edsl.exceptions.jobs import MissingRemoteInferenceError
from edsl.exceptions import MissingAPIKeyError
from edsl.jobs.buckets.BucketCollection import BucketCollection
from edsl.jobs.interviews.Interview import Interview
from edsl.language_models import LanguageModel
from edsl.results import Results
from edsl.scenarios import Scenario
from edsl.surveys import Survey

from edsl.utilities.decorators import add_edsl_version, remove_edsl_version


class Jobs(Base):
    """
    A collection of agents, scenarios and models and one survey.
    The actual running of a job is done by a `JobsRunner`, which is a subclass of `JobsRunner`.
    The `JobsRunner` is chosen by the user, and is stored in the `jobs_runner_name` attribute.
    """

    def __init__(
        self,
        survey: Survey,
        agents: Optional[list[Agent]] = None,
        models: Optional[list[LanguageModel]] = None,
        scenarios: Optional[list[Scenario]] = None,
    ):
        """Initialize a Jobs instance.

        :param survey: the survey to be used in the job
        :param agents: a list of agents
        :param models: a list of models
        :param scenarios: a list of scenarios
        """
        self.survey = survey
        self.agents = agents or []
        self.models = models or []
        self.scenarios = scenarios or []
        self.__bucket_collection = None

    def by(
        self,
        *args: Union[
            Agent,
            Scenario,
            LanguageModel,
            Sequence[Union[Agent, Scenario, LanguageModel]],
        ],
    ) -> Jobs:
        """
        Add Agents, Scenarios and LanguageModels to a job. If no objects of this type exist in the Jobs instance, it stores the new objects as a list in the corresponding attribute. Otherwise, it combines the new objects with existing objects using the object's `__add__` method.

        This 'by' is intended to create a fluent interface.

        >>> from edsl import Survey
        >>> from edsl import QuestionFreeText
        >>> q = QuestionFreeText(question_name="name", question_text="What is your name?")
        >>> j = Jobs(survey = Survey(questions=[q]))
        >>> j
        Jobs(survey=Survey(...), agents=[], models=[], scenarios=[])
        >>> from edsl import Agent; a = Agent(traits = {"status": "Sad"})
        >>> j.by(a).agents
        [Agent(traits = {'status': 'Sad'})]

        :param args: objects or a sequence (list, tuple, ...) of objects of the same type

        Notes:
        - all objects must implement the 'get_value', 'set_value', and `__add__` methods
        - agents: traits of new agents are combined with traits of existing agents. New and existing agents should not have overlapping traits, and do not increase the # agents in the instance
        - scenarios: traits of new scenarios are combined with traits of old existing. New scenarios will overwrite overlapping traits, and do not increase the number of scenarios in the instance
        - models: new models overwrite old models.
        """
        passed_objects = self._turn_args_to_list(args)

        current_objects, objects_key = self._get_current_objects_of_this_type(
            passed_objects[0]
        )

        if not current_objects:
            new_objects = passed_objects
        else:
            new_objects = self._merge_objects(passed_objects, current_objects)

        setattr(self, objects_key, new_objects)  # update the job
        return self

    def prompts(self) -> Dataset:
        """Return a Dataset of prompts that will be used.


        >>> from edsl.jobs import Jobs
        >>> Jobs.example().prompts()
        Dataset([{'interview_index': [0, 0, 1, 1, 2, 2, 3, 3]}, {'question_index': ['how_feeling', 'how_feeling_yesterday', 'how_feeling', 'how_feeling_yesterday', 'how_feeling', 'how_feeling_yesterday', 'how_feeling', 'how_feeling_yesterday']}, {'user_prompt': [Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA')]}, {'scenario_index': [Scenario({'period': 'morning'}), Scenario({'period': 'morning'}), Scenario({'period': 'afternoon'}), Scenario({'period': 'afternoon'}), Scenario({'period': 'morning'}), Scenario({'period': 'morning'}), Scenario({'period': 'afternoon'}), Scenario({'period': 'afternoon'})]}, {'system_prompt': [Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA'), Prompt(text='NA')]}])
        """

        interviews = self.interviews()
        # data = []
        interview_indices = []
        question_indices = []
        user_prompts = []
        system_prompts = []
        scenario_indices = []

        for interview_index, interview in enumerate(interviews):
            invigilators = list(interview._build_invigilators(debug=False))
            for _, invigilator in enumerate(invigilators):
                prompts = invigilator.get_prompts()
                user_prompts.append(prompts["user_prompt"])
                system_prompts.append(prompts["system_prompt"])
                interview_indices.append(interview_index)
                scenario_indices.append(invigilator.scenario)
                question_indices.append(invigilator.question.question_name)
        return Dataset(
            [
                {"interview_index": interview_indices},
                {"question_index": question_indices},
                {"user_prompt": user_prompts},
                {"scenario_index": scenario_indices},
                {"system_prompt": system_prompts},
            ]
        )

    @staticmethod
    def _turn_args_to_list(args):
        """Return a list of the first argument if it is a sequence, otherwise returns a list of all the arguments."""

        def did_user_pass_a_sequence(args):
            """Return True if the user passed a sequence, False otherwise.

            Example:

            >>> did_user_pass_a_sequence([1,2,3])
            True

            >>> did_user_pass_a_sequence(1)
            False
            """
            return len(args) == 1 and isinstance(args[0], Sequence)

        if did_user_pass_a_sequence(args):
            return list(args[0])
        else:
            return list(args)

    def _get_current_objects_of_this_type(
        self, object: Union[Agent, Scenario, LanguageModel]
    ) -> tuple[list, str]:
        """Return the current objects of the same type as the first argument.

        >>> from edsl.jobs import Jobs
        >>> j = Jobs.example()
        >>> j._get_current_objects_of_this_type(j.agents[0])
        ([Agent(traits = {'status': 'Joyful'}), Agent(traits = {'status': 'Sad'})], 'agents')
        """
        class_to_key = {
            Agent: "agents",
            Scenario: "scenarios",
            LanguageModel: "models",
        }
        for class_type in class_to_key:
            if isinstance(object, class_type) or issubclass(
                object.__class__, class_type
            ):
                key = class_to_key[class_type]
                break
        else:
            raise ValueError(
                f"First argument must be an Agent, Scenario, or LanguageModel, not {object}"
            )
        current_objects = getattr(self, key, None)
        return current_objects, key

    @staticmethod
    def _merge_objects(passed_objects, current_objects) -> list:
        """
        Combine all the existing objects with the new objects.

        For example, if the user passes in 3 agents,
        and there are 2 existing agents, this will create 6 new agents

        >>> Jobs(survey = [])._merge_objects([1,2,3], [4,5,6])
        [5, 6, 7, 6, 7, 8, 7, 8, 9]
        """
        new_objects = []
        for current_object in current_objects:
            for new_object in passed_objects:
                new_objects.append(current_object + new_object)
        return new_objects

    def interviews(self) -> list[Interview]:
        """
        Return a list of :class:`edsl.jobs.interviews.Interview` objects.

        It returns one Interview for each combination of Agent, Scenario, and LanguageModel.
        If any of Agents, Scenarios, or LanguageModels are missing, it fills in with defaults.

        >>> from edsl.jobs import Jobs
        >>> j = Jobs.example()
        >>> len(j.interviews())
        4
        >>> j.interviews()[0]
        Interview(agent = Agent(traits = {'status': 'Joyful'}), survey = Survey(...), scenario = Scenario({'period': 'morning'}), model = Model(...))
        """
        if hasattr(self, "_interviews"):
            return self._interviews
        else:
            return list(self._create_interviews())

    @classmethod
    def from_interviews(cls, interview_list):
        """Return a Jobs instance from a list of interviews."""
        survey = interview_list[0].survey
        # get all the models
        models = list(set([interview.model for interview in interview_list]))
        jobs = cls(survey)
        jobs.models = models
        jobs._interviews = interview_list
        return jobs

    def _create_interviews(self) -> Generator[Interview, None, None]:
        """
        Generate interviews.

        Note that this sets the agents, model and scenarios if they have not been set. This is a side effect of the method.
        This is useful because a user can create a job without setting the agents, models, or scenarios, and the job will still run,
        with us filling in defaults.
        """
        # if no agents, models, or scenarios are set, set them to defaults
        self.agents = self.agents or [Agent()]
        self.models = self.models or [Model()]
        # if remote, set all the models to remote
        if hasattr(self, "remote") and self.remote:
            for model in self.models:
                model.remote = True
        self.scenarios = self.scenarios or [Scenario()]
        for agent, scenario, model in product(self.agents, self.scenarios, self.models):
            yield Interview(
                survey=self.survey, agent=agent, scenario=scenario, model=model
            )

    def create_bucket_collection(self) -> BucketCollection:
        """
        Create a collection of buckets for each model.

        These buckets are used to track API calls and token usage.

        >>> from edsl.jobs import Jobs
        >>> j = Jobs.example().by(Model(temperature = 1), Model(temperature = 0.5))
        >>> bc = j.create_bucket_collection()
        >>> bc
        BucketCollection(...)
        """
        bucket_collection = BucketCollection()
        for model in self.models:
            bucket_collection.add_model(model)
        return bucket_collection

    @property
    def bucket_collection(self) -> BucketCollection:
        """Return the bucket collection. If it does not exist, create it."""
        if self.__bucket_collection is None:
            self.__bucket_collection = self.create_bucket_collection()
        return self.__bucket_collection

    def html(self):
        """Return the HTML representations for each scenario"""
        links = []
        for index, scenario in enumerate(self.scenarios):
            links.append(
                self.survey.html(
                    scenario=scenario, return_link=True, cta=f"Scenario {index}"
                )
            )
        return links

    def run(
        self,
        n: int = 1,
        debug: bool = False,
        progress_bar: bool = False,
        stop_on_exception: bool = False,
        cache: Union[Cache, bool] = None,
        remote: bool = (
            False if os.getenv("DEFAULT_RUN_MODE", "local") == "local" else True
        ),
        check_api_keys: bool = False,
        sidecar_model: Optional[LanguageModel] = None,
        batch_mode: Optional[bool] = None,
        print_exceptions=False,
    ) -> Results:
        """
        Runs the Job: conducts Interviews and returns their results.

        :param n: how many times to run each interview
        :param debug: prints debug messages
        :param verbose: prints messages
        :param progress_bar: shows a progress bar
        :param stop_on_exception: stops the job if an exception is raised
        :param cache: a cache object to store results
        :param remote: run the job remotely
        :param check_api_keys: check if the API keys are valid
        :batch_mode: run the job in batch mode i.e., no expecation of interaction with the user

        """
        if batch_mode is not None:
            raise NotImplementedError(
                "Batch mode is deprecated. Please update your code to not include 'batch_mode' in the 'run' method."
            )

        self.remote = remote

        if self.remote:
            ## TODO: This should be a coop check
            if os.getenv("EXPECTED_PARROT_API_KEY", None) is None:
                raise MissingRemoteInferenceError()

        if not self.remote:
            if check_api_keys:
                for model in self.models + [Model()]:
                    if not model.has_valid_api_key():
                        raise MissingAPIKeyError(
                            model_name=str(model.model),
                            inference_service=model._inference_service_,
                        )

        # handle cache
        if cache is None:
            cache = CacheHandler().get_cache()
        if cache is False:
            cache = Cache()

        results = self._run_local(
            n=n,
            debug=debug,
            progress_bar=progress_bar,
            cache=cache,
            stop_on_exception=stop_on_exception,
            sidecar_model=sidecar_model,
            print_exceptions=print_exceptions,
        )
        results.cache = cache.new_entries_cache()

        return results

    def _run_local(self, *args, **kwargs):
        """Run the job locally."""
        from edsl.jobs.runners.JobsRunnerAsyncio import JobsRunnerAsyncio

        results = JobsRunnerAsyncio(self).run(*args, **kwargs)
        return results

    #######################
    # Dunder methods
    #######################
    def print(self):
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()))

    def __repr__(self) -> str:
        """Return an eval-able string representation of the Jobs instance."""
        return f"Jobs(survey={repr(self.survey)}, agents={repr(self.agents)}, models={repr(self.models)}, scenarios={repr(self.scenarios)})"

    def _repr_html_(self) -> str:
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()))

    def __len__(self) -> int:
        """Return the maximum number of questions that will be asked while running this job.
        Note that this is the maximum number of questions, not the actual number of questions that will be asked, as some questions may be skipped.

        >>> from edsl.jobs import Jobs
        >>> len(Jobs.example())
        8
        """
        number_of_questions = (
            len(self.agents or [1])
            * len(self.scenarios or [1])
            * len(self.models or [1])
            * len(self.survey)
        )
        return number_of_questions

    #######################
    # Serialization methods
    #######################
    @add_edsl_version
    def to_dict(self) -> dict:
        """Convert the Jobs instance to a dictionary."""
        return {
            "survey": self.survey.to_dict(),
            "agents": [agent.to_dict() for agent in self.agents],
            "models": [model.to_dict() for model in self.models],
            "scenarios": [scenario.to_dict() for scenario in self.scenarios],
        }

    @classmethod
    @remove_edsl_version
    def from_dict(cls, data: dict) -> Jobs:
        """Creates a Jobs instance from a dictionary."""
        return cls(
            survey=Survey.from_dict(data["survey"]),
            agents=[Agent.from_dict(agent) for agent in data["agents"]],
            models=[LanguageModel.from_dict(model) for model in data["models"]],
            scenarios=[Scenario.from_dict(scenario) for scenario in data["scenarios"]],
        )

    def __eq__(self, other: Jobs) -> bool:
        """Return True if the Jobs instance is equal to another Jobs instance."""
        return self.to_dict() == other.to_dict()

    #######################
    # Example methods
    #######################
    @classmethod
    def example(cls, throw_exception_probability=0) -> Jobs:
        """Return an example Jobs instance.

        :param throw_exception_probability: the probability that an exception will be thrown when answering a question. This is useful for testing error handling.

        >>> Jobs.example()
        Jobs(...)

        """
        import random
        from edsl.questions import QuestionMultipleChoice
        from edsl import Agent

        # (status, question, period)
        agent_answers = {
            ("Joyful", "how_feeling", "morning"): "OK",
            ("Joyful", "how_feeling", "afternoon"): "Great",
            ("Joyful", "how_feeling_yesterday", "morning"): "Great",
            ("Joyful", "how_feeling_yesterday", "afternoon"): "Good",
            ("Sad", "how_feeling", "morning"): "Terrible",
            ("Sad", "how_feeling", "afternoon"): "OK",
            ("Sad", "how_feeling_yesterday", "morning"): "OK",
            ("Sad", "how_feeling_yesterday", "afternoon"): "Terrible",
        }

        def answer_question_directly(self, question, scenario):
            """Return the answer to a question. This is a method that can be added to an agent."""

            if random.random() < throw_exception_probability:
                raise Exception("Error!")
            return agent_answers[
                (self.traits["status"], question.question_name, scenario["period"])
            ]

        sad_agent = Agent(traits={"status": "Sad"})
        joy_agent = Agent(traits={"status": "Joyful"})

        sad_agent.add_direct_question_answering_method(answer_question_directly)
        joy_agent.add_direct_question_answering_method(answer_question_directly)

        q1 = QuestionMultipleChoice(
            question_text="How are you this {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling",
        )
        q2 = QuestionMultipleChoice(
            question_text="How were you feeling yesterday {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling_yesterday",
        )
        base_survey = Survey(questions=[q1, q2])

        job = base_survey.by(
            Scenario({"period": "morning"}), Scenario({"period": "afternoon"})
        ).by(joy_agent, sad_agent)

        return job

    def rich_print(self):
        """Print a rich representation of the Jobs instance."""
        from rich.table import Table

        table = Table(title="Jobs")
        table.add_column("Jobs")
        table.add_row(self.survey.rich_print())
        return table

    def code(self):
        """Return the code to create this instance."""
        raise NotImplementedError


def main():
    """Run the module's doctests."""
    from edsl.jobs import Jobs
    from edsl.data.Cache import Cache

    job = Jobs.example()
    len(job) == 8
    results = job.run(debug=True, cache=Cache())
    len(results) == 8
    results


if __name__ == "__main__":
    """Run the module's doctests."""
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    # from edsl.jobs import Jobs

    # job = Jobs.example()
    # len(job) == 8
    # results, info = job.run(debug=True)
    # len(results) == 8
    # results



import json
import asyncio

from edsl.utilities.utilities import clean_json


async def async_repair(bad_json, error_message=""):
    s = clean_json(bad_json)
    from edsl import Model

    m = Model()

    try:
        # this is the OpenAI version, but that's fine
        valid_dict = json.loads(s)
        success = True
    except json.JSONDecodeError:
        valid_dict = {}
        success = False
        # print("Replacing control characters didn't work. Trying with the model.")
    else:
        return valid_dict, success

    prompt = f"""This is the output from a less capable language model.  
    It was supposed to respond with just a JSON object with an answer to a question and some commentary, 
    in a field called "comment" next to "answer".
    Please repair this bad JSON: {bad_json}."""

    if error_message:
        prompt += f" Parsing error message: {error_message}"

    try:
        results = await m.async_execute_model_call(
            prompt,
            system_prompt="You are a helpful agent. Only return the repaired JSON, nothing else.",
        )
    except Exception as e:
        return {}, False

    try:
        # this is the OpenAI version, but that's fine
        valid_dict = json.loads(results["choices"][0]["message"]["content"])
        success = True
    except json.JSONDecodeError:
        valid_dict = {}
        success = False

    return valid_dict, success


def repair_wrapper(bad_json, error_message=""):
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # Add repair as a task to the running loop
            task = loop.create_task(async_repair(bad_json, error_message))
            return task
        else:
            # Run a new event loop for repair
            return loop.run_until_complete(async_repair(bad_json, error_message))
    except RuntimeError:
        # Create a new event loop if one is not already available
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(async_repair(bad_json, error_message))


def repair(bad_json, error_message=""):
    return repair_wrapper(bad_json, error_message)


# Example usage:
# result, success = repair_wrapper('{"name": "John Doe", "age": 30,}')  # example bad JSON


# def repair_wrapper(bad_json, error_message=""):
#     loop = asyncio.get_event_loop()
#     if loop.is_running():
#         # Add repair as a task to the running loop
#         task = loop.create_task(repair(bad_json, error_message))
#         return task
#     else:
#         # Run a new event loop for repair
#         return loop.run_until_complete(repair(bad_json, error_message))


# Example usage:
# result, success = repair_wrapper('{"name": "John Doe", "age": 30,}')  # example bad JSON


if __name__ == "__main__":
    bad_json = """
    {
      'answer': "The problematic phrase in the excerpt is \'typically\'. This word is vague and can lead to different interpretations. An alternative phrasing that would be less problematic is: 
      'On average, how long do you cook scrambled eggs?}
    """
    try:
        json.loads(bad_json)
        print("Loaded")
    except json.JSONDecodeError as e:
        error_message = str(e)
        repaired, success = repair(bad_json, error_message)
        print(f"Repaired: {repaired}")



"""Mixin class for exporting results."""

import base64
import csv
import io
import random
from functools import wraps

from typing import Literal, Optional

from edsl.utilities.utilities import is_notebook

from IPython.display import HTML, display
import pandas as pd
from edsl.utilities.interface import (
    print_dataset_with_rich,
    print_list_of_dicts_as_html_table,
    print_list_of_dicts_as_markdown_table,
)


class ResultsExportMixin:
    """Mixin class for exporting Results objects."""

    def _convert_decorator(func):
        """Convert the Results object to a Dataset object before calling the function."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            """Return the function with the Results object converted to a Dataset object."""
            if self.__class__.__name__ == "Results":
                return func(self.select(), *args, **kwargs)
            elif self.__class__.__name__ == "Dataset":
                return func(self, *args, **kwargs)
            else:
                raise Exception(
                    f"Class {self.__class__.__name__} not recognized as a Results or Dataset object."
                )

        return wrapper

    @_convert_decorator
    def relevant_columns(
        self, data_type: Optional[str] = None, remove_prefix=False
    ) -> list:
        """Return the set of keys that are present in the dataset.

        >>> d = Dataset([{'a.b':[1,2,3,4]}])
        >>> d.relevant_columns()
        ['a.b']

        >>> d.relevant_columns(remove_prefix=True)
        ['b']

        >>> from edsl.results import Results; Results.example().select('how_feeling', 'how_feeling_yesterday').relevant_columns()
        ['answer.how_feeling', 'answer.how_feeling_yesterday']
        """
        columns = [list(x.keys())[0] for x in self]
        # columns = set([list(result.keys())[0] for result in self.data])
        if remove_prefix:
            columns = [column.split(".")[-1] for column in columns]

        if data_type:
            columns = [
                column for column in columns if column.split(".")[0] == data_type
            ]

        return columns

    # @_convert_decorator
    def sample(self, n: int) -> "Results":
        """Return a random sample of the results.

        :param n: The number of samples to return.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> len(r.sample(2))
        2
        """
        indices = None

        for entry in self:
            key, values = list(entry.items())[0]
            if indices is None:  # gets the indices for the first time
                indices = list(range(len(values)))
                sampled_indices = random.sample(indices, n)
                if n > len(indices):
                    raise ValueError(
                        f"Cannot sample {n} items from a list of length {len(indices)}."
                    )
            entry[key] = [values[i] for i in sampled_indices]

        return self

    @_convert_decorator
    def _make_tabular(self, remove_prefix: bool, pretty_labels: Optional[dict] = None):
        """Turn the results into a tabular format.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling')._make_tabular(remove_prefix = True)
        (['how_feeling'], [['OK'], ['Great'], ['Terrible'], ['OK']])

        >>> r.select('how_feeling')._make_tabular(remove_prefix = True, pretty_labels = {'how_feeling': "How are you feeling"})
        (['How are you feeling'], [['OK'], ['Great'], ['Terrible'], ['OK']])
        """
        d = {}
        full_header = sorted(list(self.relevant_columns()))
        for entry in self.data:
            key, list_of_values = list(entry.items())[0]
            d[key] = list_of_values
        if remove_prefix:
            header = [h.split(".")[-1] for h in full_header]
        else:
            header = full_header
        num_observations = len(list(self[0].values())[0])
        rows = []
        # rows.append(header)
        for i in range(num_observations):
            row = [d[h][i] for h in full_header]
            rows.append(row)
        if pretty_labels is not None:
            header = [pretty_labels.get(h, h) for h in header]
        return header, rows

    def print_long(self, max_rows=None) -> None:
        """Print the results in long format.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').print_long(max_rows = 2)
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━┓
        ┃ Result index ┃ Key         ┃ Value ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━╇━━━━━━━┩
        │ 0            │ how_feeling │ OK    │
        │ 1            │ how_feeling │ Great │
        └──────────────┴─────────────┴───────┘
        """
        from edsl.utilities.interface import print_results_long

        print_results_long(self, max_rows=max_rows)

    @_convert_decorator
    def print(
        self,
        pretty_labels: Optional[dict] = None,
        filename: Optional[str] = None,
        format: Literal["rich", "html", "markdown"] = None,
        interactive: bool = False,
        split_at_dot: bool = True,
        max_rows=None,
        iframe: Optional[bool] = None,
    ) -> None:
        """Print the results in a pretty format.

        :param pretty_labels: A dictionary of pretty labels for the columns.
        :param filename: The filename to save the results to.
        :param format: The format to print the results in. Options are 'rich', 'html', or 'markdown'.
        :param interactive: Whether to print the results interactively in a Jupyter notebook.
        :param split_at_dot: Whether to split the column names at the last dot w/ a newline.

        Example: Print in rich format at the terminal

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').print(format = "rich")
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ OK           │
        ├──────────────┤
        │ Great        │
        ├──────────────┤
        │ Terrible     │
        ├──────────────┤
        │ OK           │
        └──────────────┘

        Example: using the pretty_labels parameter

        >>> r.select('how_feeling').print(format="rich", pretty_labels = {'answer.how_feeling': "How are you feeling"})
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ How are you feeling ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ OK                  │
        ├─────────────────────┤
        │ Great               │
        ├─────────────────────┤
        │ Terrible            │
        ├─────────────────────┤
        │ OK                  │
        └─────────────────────┘

        Example: printing in markdown format

        >>> r.select('how_feeling').print(format='markdown')
        | answer.how_feeling |
        |--|
        | OK |
        | Great |
        | Terrible |
        | OK |
        ...
        """
        if format is None:
            if is_notebook():
                format = "html"
            else:
                format = "rich"

        if pretty_labels is None:
            pretty_labels = {}

        iframe = iframe or False

        if format not in ["rich", "html", "markdown"]:
            raise ValueError("format must be one of 'rich', 'html', or 'markdown'.")

        new_data = []
        for index, entry in enumerate(self):
            key, list_of_values = list(entry.items())[0]
            new_data.append({pretty_labels.get(key, key): list_of_values})

        if max_rows is not None:
            for entry in new_data:
                for key in entry:
                    actual_rows = len(entry[key])
                    entry[key] = entry[key][:max_rows]
            # print(f"Showing only the first {max_rows} rows of {actual_rows} rows.")

        if format == "rich":
            print_dataset_with_rich(
                new_data, filename=filename, split_at_dot=split_at_dot
            )
        elif format == "html":
            notebook = is_notebook()
            html_source = print_list_of_dicts_as_html_table(
                new_data, filename=None, interactive=interactive, notebook=notebook
            )
            if iframe:
                import html

                height = 200
                width = 600
                escaped_output = html.escape(html_source)
                # escaped_output = html_source
                iframe = f""""
                <iframe srcdoc="{ escaped_output }" style="width: {width}px; height: {height}px;"></iframe>
                """
                display(HTML(iframe))
            else:
                display(HTML(html_source))
        elif format == "markdown":
            print_list_of_dicts_as_markdown_table(new_data, filename=filename)

    @_convert_decorator
    def to_csv(
        self,
        filename: Optional[str] = None,
        remove_prefix: bool = False,
        download_link: bool = False,
        pretty_labels: Optional[dict] = None,
    ):
        """Export the results to a CSV file.

        :param filename: The filename to save the CSV file to.
        :param remove_prefix: Whether to remove the prefix from the column names.
        :param download_link: Whether to display a download link in a Jupyter notebook.

        Example:

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_csv()
        'answer.how_feeling\\r\\nOK\\r\\nGreat\\r\\nTerrible\\r\\nOK\\r\\n'
        """
        if pretty_labels is None:
            pretty_labels = {}
        header, rows = self._make_tabular(
            remove_prefix=remove_prefix, pretty_labels=pretty_labels
        )

        if filename is not None:
            with open(filename, "w") as f:
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows(rows)
        else:
            output = io.StringIO()
            writer = csv.writer(output)
            writer.writerow(header)
            writer.writerows(rows)

            if download_link:
                csv_file = output.getvalue()
                b64 = base64.b64encode(csv_file.encode()).decode()
                download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
                display(HTML(download_link))
            else:
                return output.getvalue()

    @_convert_decorator
    def to_pandas(self, remove_prefix: bool = False) -> pd.DataFrame:
        """Convert the results to a pandas DataFrame.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_pandas()
          answer.how_feeling
        0                 OK
        1              Great
        2           Terrible
        3                 OK
        """
        csv_string = self.to_csv(remove_prefix=remove_prefix)
        csv_buffer = io.StringIO(csv_string)
        df = pd.read_csv(csv_buffer)
        df_sorted = df.sort_index(axis=1)  # Sort columns alphabetically
        return df_sorted

    @_convert_decorator
    def to_scenario_list(self, remove_prefix: bool = True) -> list[dict]:
        """Convert the results to a list of dictionaries, one per scenario.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_scenario_list()
        ScenarioList([Scenario({'how_feeling': 'OK'}), Scenario({'how_feeling': 'Great'}), Scenario({'how_feeling': 'Terrible'}), Scenario({'how_feeling': 'OK'})])
        """
        from edsl import ScenarioList, Scenario

        list_of_dicts = self.to_dicts(remove_prefix=remove_prefix)
        return ScenarioList([Scenario(d) for d in list_of_dicts])

    def to_agent_list(self, remove_prefix: bool = True):
        from edsl import AgentList, Agent

        list_of_dicts = self.to_dicts(remove_prefix=remove_prefix)
        return AgentList([Agent(d) for d in list_of_dicts])

    @_convert_decorator
    def to_dicts(self, remove_prefix: bool = True) -> list[dict]:
        """Convert the results to a list of dictionaries.

        :param remove_prefix: Whether to remove the prefix from the column names.

        >>> from edsl.results import Results
        >>> r = Results.example()
        >>> r.select('how_feeling').to_dicts()
        [{'how_feeling': 'OK'}, {'how_feeling': 'Great'}, {'how_feeling': 'Terrible'}, {'how_feeling': 'OK'}]

        """
        list_of_keys = []
        list_of_values = []
        for entry in self:
            key, values = list(entry.items())[0]
            list_of_keys.append(key)
            list_of_values.append(values)

        if remove_prefix:
            list_of_keys = [key.split(".")[-1] for key in list_of_keys]
        #        else:
        #            list_of_keys = [key.replace(".", "_") for key in list_of_keys]

        list_of_dicts = []
        for entries in zip(*list_of_values):
            list_of_dicts.append(dict(zip(list_of_keys, entries)))

        return list_of_dicts

    @_convert_decorator
    def to_list(self, flatten=False, remove_none=False) -> list[list]:
        """Convert the results to a list of lists.

        >>> from edsl.results import Results
        >>> Results.example().select('how_feeling', 'how_feeling_yesterday')
        Dataset([{'answer.how_feeling': ['OK', 'Great', 'Terrible', 'OK']}, {'answer.how_feeling_yesterday': ['Great', 'Good', 'OK', 'Terrible']}])

        >>> Results.example().select('how_feeling', 'how_feeling_yesterday').to_list()
        [('OK', 'Great'), ('Great', 'Good'), ('Terrible', 'OK'), ('OK', 'Terrible')]

        >>> r = Results.example()
        >>> r.select('how_feeling').to_list()
        ['OK', 'Great', 'Terrible', 'OK']
        """
        if len(self.relevant_columns()) > 1 and flatten:
            raise ValueError(
                "Cannot flatten a list of lists when there are multiple columns selected."
            )

        if len(self.relevant_columns()) == 1:
            # if only one 'column' is selected (which is typical for this method
            list_to_return = list(self[0].values())[0]
        else:
            keys = self.relevant_columns()
            data = self.to_dicts(remove_prefix=False)
            list_to_return = []
            for d in data:
                list_to_return.append(tuple([d[key] for key in keys]))

        if remove_none:
            list_to_return = [item for item in list_to_return if item is not None]

        if flatten:
            new_list = []
            for item in list_to_return:
                if isinstance(item, list):
                    new_list.extend(item)
                else:
                    new_list.append(item)
            list_to_return = new_list

        return list_to_return

    @_convert_decorator
    def html(
        self, filename: str = None, cta: str = "Open in browser", return_link=False
    ):
        import os
        import tempfile

        df = self.to_pandas()

        if filename is None:
            current_directory = os.getcwd()
            filename = tempfile.NamedTemporaryFile(
                "w", delete=False, suffix=".html", dir=current_directory
            ).name

        with open(filename, "w") as f:
            f.write(df.to_html())

        if is_notebook():
            html_url = f"/files/{filename}"
            html_link = f'<a href="{html_url}" target="_blank">{cta}</a>'
            display(HTML(html_link))
        else:
            print(f"Saved to {filename}")
            import webbrowser
            import os

            webbrowser.open(f"file://{os.path.abspath(filename)}")
            # webbrowser.open(filename)

        if return_link:
            return filename

    @_convert_decorator
    def tally(self, *fields: Optional[str], top_n=None, format=None):
        """Tally the values of a field or perform a cross-tab of multiple fields.

        :param fields: The field(s) to tally, multiple fields for cross-tabulation.

        >>> r = Results.example()
        >>> r.select('how_feeling').tally('answer.how_feeling')
        {'OK': 2, 'Great': 1, 'Terrible': 1}
        >>> r.tally('field1', 'field2')
        {('X', 'A'): 1, ('X', 'B'): 1, ('Y', 'A'): 1}
        """
        from collections import Counter

        if len(fields) == 0:
            fields = self.relevant_columns()

        relevant_columns_without_prefix = [
            column.split(".")[-1] for column in self.relevant_columns()
        ]

        if not all(
            f in self.relevant_columns() or f in relevant_columns_without_prefix
            for f in fields
        ):
            raise ValueError("One or more specified fields are not in the dataset.")

        if len(fields) == 1:
            field = fields[0]
            values = self._key_to_value(field)
        else:
            values = list(zip(*(self._key_to_value(field) for field in fields)))

        tally = dict(Counter(values))
        sorted_tally = dict(sorted(tally.items(), key=lambda item: -item[1]))
        if top_n is not None:
            sorted_tally = dict(list(sorted_tally.items())[:top_n])

        if format is not None:
            if format == "rich":
                from edsl.utilities.interface import print_tally_with_rich

                print_tally_with_rich(sorted_tally)
                return None

        return sorted_tally


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations
from abc import ABCMeta

from edsl.enums import QuestionType
from edsl.exceptions.questions import QuestionMissingTypeError, QuestionBadTypeError


class RegisterQuestionsMeta(ABCMeta):
    """Metaclass to register output elements in a registry i.e., those that have a parent."""

    _registry = {}  # Initialize the registry as a dictionary

    def __init__(cls, name, bases, dct):
        """Initialize the class and adds it to the registry if it's not the base class."""
        super(RegisterQuestionsMeta, cls).__init__(name, bases, dct)
        if name != "QuestionBase":
            ## Enforce that all questions have a question_type class attribute
            ## and it comes from our enum of valid question types.
            if not hasattr(cls, "question_type"):
                raise QuestionMissingTypeError(
                    "Question must have a question_type class attribute"
                )

            if not QuestionType.is_value_valid(cls.question_type):
                acceptable_values = [item.value for item in QuestionType]
                raise QuestionBadTypeError(
                    f"""question_type must be one of {QuestionType} values, which are 
                                currently {acceptable_values}"""
                    ""
                )

            RegisterQuestionsMeta._registry[name] = cls

    @classmethod
    def get_registered_classes(cls):
        """Return the registry of registered classes."""
        return cls._registry

    @classmethod
    def question_types_to_classes(
        cls,
    ):
        """Return a dictionary of question types to classes."""
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "question_type"):
                d[cls.question_type] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a question_type class attribute"
                )
        return d



import re
from textwrap import dedent


class GeneralErrors(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return self.make_urls_clickable(self.message)

    @staticmethod
    def make_urls_clickable(text):
        url_pattern = r"https?://[^\s]+"
        urls = re.findall(url_pattern, text)
        for url in urls:
            clickable_url = f"\033]8;;{url}\007{url}\033]8;;\007"
            text = text.replace(url, clickable_url)
        return text


class MissingAPIKeyError(GeneralErrors):
    def __init__(self, model_name, inference_service):
        full_message = dedent(
            f"""
        An API Key for model `{model_name}` is missing from the .env file.
        This key is associated with the inference service `{inference_service}`.
        Please see https://docs.expectedparrot.com/en/latest/api_keys.html for more information.
        """
        )
        super().__init__(full_message)



"""Mixin with validators for LLM answers to questions."""

import re
from typing import Any, Type, Union
from edsl.exceptions import (
    QuestionAnswerValidationError,
)


class AnswerValidatorMixin:
    """
    Mixin with validators for LLM answers to questions.

    - Template validation: validators for the entire answer object format
    - Value validation: validators for specific values
    - Question specific validation: validators for specific question types
    """

    #####################
    # TEMPLATE VALIDATION
    #####################
    def _validate_answer_template_basic(self, answer: Any) -> None:
        """Check that the answer (i) is a dictionary (ii) has an 'answer' key.

        >>> avm = AnswerValidatorMixin()
        >>> avm._validate_answer_template_basic({'answer': 1})
        >>> avm._validate_answer_template_basic([])
        Traceback (most recent call last):
        ...
        edsl.exceptions.questions.QuestionAnswerValidationError: Answer must be a dictionary (got []).

        - E.g., both {'answer': 1} and {'answer': {'a': 1}, 'other_key'=[1,2,3]} are valid
        """
        if not isinstance(answer, dict):
            raise QuestionAnswerValidationError(
                f"Answer must be a dictionary (got {answer})."
            )
        if not "answer" in answer:
            raise QuestionAnswerValidationError(
                f"Answer must have an 'answer' key (got {answer})."
            )

    #####################
    # VALUE VALIDATION
    #####################
    def _validate_answer_key_value(
        self, answer: dict[str, Any], key: str, of_type: Type
    ) -> None:
        """Check that the value of a key is of the specified type."""
        if not isinstance(answer.get(key), of_type):
            raise QuestionAnswerValidationError(
                f"""Answer key '{key}' must be of type {of_type.__name__};
                (got {answer.get(key)}) which is of type {type(answer.get(key))}."""
            )

    def _validate_answer_key_value_numeric(
        self, answer: dict[str, Any], key: str
    ) -> None:
        """Check that the value of a key is numeric (int or float).

        >>> avm = AnswerValidatorMixin()
        >>> avm._validate_answer_key_value_numeric({'answer': 1}, 'answer')
        >>> avm._validate_answer_key_value_numeric({'answer': 'poo'}, 'answer')
        Traceback (most recent call last):
        ...
        edsl.exceptions.questions.QuestionAnswerValidationError: Answer should be numerical (int or float). Got 'poo'
        """
        value = answer.get(key)
        initial_value = value
        if type(value) == str:
            value = value.replace(",", "")
            value = "".join(re.findall(r"[-+]?\d*\.\d+|\d+", value))
            if value.isdigit():
                value = int(value)
            else:
                try:
                    float(value)
                    value = float(value)
                except ValueError:
                    raise QuestionAnswerValidationError(
                        f"Answer should be numerical (int or float). Got '{initial_value}'"
                    )
            return None
        elif type(value) == int or type(value) == float:
            return None
        else:
            raise QuestionAnswerValidationError(
                f"Answer should be numerical (int or float)."
            )

    #####################
    # QUESTION SPECIFIC VALIDATION
    #####################
    def _validate_answer_budget(self, answer: dict[str, Any]) -> None:
        """Validate QuestionBudget-specific answer.

        Check that answer["answer"]:
        - has keys that are in the range of the number of options
        - has values that are non-negative integers
        - has values that sum to `budget_sum`
        - contains all keys in the range of the number of options
        """
        answer = answer.get("answer")
        budget_sum = self.budget_sum
        acceptable_answer_keys = set(range(len(self.question_options)))
        answer_keys = set([int(k) for k in answer.keys()])
        current_sum = sum(answer.values())
        if not current_sum == budget_sum:
            raise QuestionAnswerValidationError(
                f"Budget sum must be {budget_sum}, but got {current_sum}."
            )
        if any(v < 0 for v in answer.values()):
            raise QuestionAnswerValidationError(
                f"Budget values must be positive, but got {answer_keys}."
            )
        if any([int(key) not in acceptable_answer_keys for key in answer.keys()]):
            raise QuestionAnswerValidationError(
                f"Budget keys must be in {acceptable_answer_keys}, but got {answer_keys}."
            )
        if acceptable_answer_keys != answer_keys:
            missing_keys = acceptable_answer_keys - answer_keys
            raise QuestionAnswerValidationError(
                f"All but keys must be represented in the answer. Missing: {missing_keys}."
            )

    def _validate_answer_checkbox(self, answer: dict[str, Union[str, int]]) -> None:
        """Validate QuestionCheckbox-specific answer.

        :param answer: Answer to validate

        >>> avm = AnswerValidatorMixin()
        >>> avm.question_options = ["a", "b", "c"]
        >>> avm.min_selections = 1
        >>> avm.max_selections = 2
        >>> avm._validate_answer_checkbox({"answer": ["0", "1"]})
        >>> avm._validate_answer_checkbox({"answer": []})
        Traceback (most recent call last):
        ...
        edsl.exceptions.questions.QuestionAnswerValidationError:...

        Check that answer["answer"]:
        - has elements that are strings, bytes-like objects or real numbers evaluating to integers
        - has elements that are in the range of the number of options
        - has at least `min_selections` elements, if provided
        - has at most `max_selections` elements, if provided
        """
        answer_codes = answer["answer"]
        try:
            answer_codes = [int(k) for k in answer["answer"]]
        except:
            raise QuestionAnswerValidationError(
                f"Answer codes must be a list of strings, bytes-like objects or real numbers (got {answer['answer']})."
            )
        acceptable_values = list(range(len(self.question_options)))
        for answer_code in answer_codes:
            if answer_code not in acceptable_values:
                raise QuestionAnswerValidationError(
                    f"Answer code {answer_code} has elements not in {acceptable_values}."
                )
        if self.min_selections is not None and len(answer_codes) < self.min_selections:
            raise QuestionAnswerValidationError(
                f"Answer codes, {answer_codes}, has fewer than {self.min_selections} options selected."
            )
        if self.max_selections is not None and len(answer_codes) > self.max_selections:
            raise QuestionAnswerValidationError(
                f"Answer codes, {answer_codes}, has more than {self.max_selections} options selected."
            )

    def _validate_answer_extract(self, answer: dict[str, Any]) -> None:
        """Validate QuestionExtract-specific answer.

        Check that answer["answer"]:
        - does not have keys that are not in the answer template
        - has all keys that are in the answer template
        """
        value = answer.get("answer")
        acceptable_answer_keys = set(self.answer_template.keys())
        if any([key not in acceptable_answer_keys for key in value.keys()]):
            raise QuestionAnswerValidationError(
                f"Answer keys must be in {acceptable_answer_keys}, but got {value.keys()}."
            )
        if any([key not in value.keys() for key in acceptable_answer_keys]):
            raise QuestionAnswerValidationError(
                f"Answer must have all keys in {acceptable_answer_keys}, but got {value.keys()}."
            )

    def _validate_answer_list(self, answer: dict[str, Union[list, str]]) -> None:
        """Validate QuestionList-specific answer.

        Check that answer["answer"]:
        - is not empty, if `allow_nonresponse` is False
        - has no more than `max_list_items` elements
        - has no empty strings
        """
        value = answer.get("answer")
        if (
            hasattr(self, "allow_nonresponse")
            and self.allow_nonresponse == False
            and (value == [] or value is None)
        ):
            raise QuestionAnswerValidationError("You must provide a response.")

        if (
            hasattr(self, "max_list_items")
            and self.max_list_items is not None
            and (len(value) > self.max_list_items)
        ):
            raise QuestionAnswerValidationError("Response has too many items.")

        if any([item == "" for item in value]):
            raise QuestionAnswerValidationError(
                f"Answer cannot contain empty strings, but got {value}."
            )

    def _validate_answer_numerical(self, answer: dict) -> None:
        """Validate QuestionNumerical-specific answer.

        Check that answer["answer"]:
        - is not less than `min_value`
        - is not greater than `max_value`
        """
        value = float(answer.get("answer"))
        if self.min_value is not None and value < self.min_value:
            raise QuestionAnswerValidationError(
                f"Value {value} is less than {self.min_value}"
            )
        if self.max_value is not None and value > self.max_value:
            raise QuestionAnswerValidationError(
                f"Value {value} is greater than {self.max_value}"
            )
        return value

    def _validate_answer_multiple_choice(
        self, answer: dict[str, Union[str, int]]
    ) -> None:
        """Validate QuestionMultipleChoice-specific answer.

        Check that answer["answer"]:
        - is a string, bytes-like object or real number
        - is a non-negative integer
        - is in the range of the number of options
        """
        try:
            value = int(answer.get("answer"))
        except:
            raise QuestionAnswerValidationError(
                f"Answer code must be a string, a bytes-like object or a real number (got {answer.get('answer')})."
            )
        if not value >= 0:
            raise QuestionAnswerValidationError(
                f"Answer code must be a non-negative integer (got {value})."
            )
        if int(value) not in range(len(self.question_options)):
            raise QuestionAnswerValidationError(
                f"Answer code {value} must be in {list(range(len(self.question_options)))}."
            )

    def _validate_answer_rank(self, answer: dict[str, Union[str, int]]) -> None:
        """Validate QuestionRank-specific answer.

        Check that answer["answer"]:
        - contains only integers
        - contains only integers in the range of the number of options
        - has the correct number of elements
        """
        value = answer.get("answer")
        acceptable_values = list(range(len(self.question_options)))
        for v in value:
            try:
                answer_code = int(v)
            except ValueError:
                raise QuestionAnswerValidationError(
                    f"Rank answer {value} has elements that are not integers, namely {v}."
                )
            except TypeError:
                raise QuestionAnswerValidationError(
                    f"Rank answer {value} has elements that are not integers, namely {v}."
                )
            if answer_code not in acceptable_values:
                raise QuestionAnswerValidationError(
                    f"Answer {value} has elements not in {acceptable_values}, namely {v}."
                )
        if len(value) != self.num_selections:
            raise QuestionAnswerValidationError(
                f"Rank answer {value}, but exactly {self.num_selections} selections required."
            )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl import Agent, AgentList, Cache, Jobs, Results, Scenario, ScenarioList, Survey
from edsl.questions import QuestionBase
from typing import Literal, Type, Union

EDSLObject = Union[
    Agent,
    AgentList,
    Cache,
    Jobs,
    Type[QuestionBase],
    Results,
    Scenario,
    ScenarioList,
    Survey,
]

ObjectType = Literal[
    "agent",
    "agent_list",
    "cache",
    "job",
    "question",
    "results",
    "scenario",
    "scenario_list",
    "survey",
]

ObjectPage = Literal[
    "agents",
    "agentlists",
    "caches",
    "jobs",
    "questions",
    "results",
    "scenarios",
    "scenariolists",
    "surveys",
]

VisibilityType = Literal[
    "private",
    "public",
    "unlisted",
]


class ObjectRegistry:
    """
    Utility class to map object types to database models.
    """

    objects = [
        {
            "object_type": "agent",
            "edsl_class": Agent,
            "object_page": "agents",
        },
        {
            "object_type": "agent_list",
            "edsl_class": AgentList,
            "object_page": "agentlists",
        },
        {
            "object_type": "cache",
            "edsl_class": Cache,
            "object_page": "caches",
        },
        {
            "object_type": "job",
            "edsl_class": Jobs,
            "object_page": "jobs",
        },
        {
            "object_type": "question",
            "edsl_class": QuestionBase,
            "object_page": "questions",
        },
        {
            "object_type": "results",
            "edsl_class": Results,
            "object_page": "results",
        },
        {
            "object_type": "scenario",
            "edsl_class": Scenario,
            "object_page": "scenarios",
        },
        {
            "object_type": "scenario_list",
            "edsl_class": ScenarioList,
            "object_page": "scenariolists",
        },
        {
            "object_type": "survey",
            "edsl_class": Survey,
            "object_page": "surveys",
        },
    ]
    object_type_to_edsl_class = {o["object_type"]: o["edsl_class"] for o in objects}
    edsl_class_to_object_type = {
        o["edsl_class"].__name__: o["object_type"] for o in objects
    }
    object_type_to_object_page = {o["object_type"]: o["object_page"] for o in objects}

    @classmethod
    def get_object_type_by_edsl_class(cls, edsl_object: EDSLObject) -> ObjectType:
        if isinstance(edsl_object, type):
            edsl_class_name = edsl_object.__name__
        else:
            edsl_class_name = type(edsl_object).__name__
        if edsl_class_name.startswith("Question"):
            edsl_class_name = "QuestionBase"
        object_type = cls.edsl_class_to_object_type.get(edsl_class_name)
        if object_type is None:
            raise ValueError(f"Object type not found for {edsl_object=}")
        return object_type

    @classmethod
    def get_edsl_class_by_object_type(cls, object_type: ObjectType) -> EDSLObject:
        EDSL_object = cls.object_type_to_edsl_class.get(object_type)
        if EDSL_object is None:
            raise ValueError(f"EDSL class not found for {object_type=}")
        return EDSL_object

    @classmethod
    def get_object_page_by_object_type(cls, object_type: ObjectType) -> ObjectPage:
        return cls.object_type_to_object_page.get(object_type)



import aiohttp
import json
import os
import requests
from typing import Any, Optional, Union, Literal
from uuid import UUID
import edsl
from edsl import CONFIG, CacheEntry
from edsl.coop.utils import EDSLObject, ObjectRegistry, ObjectType, VisibilityType


class Coop:
    """
    Client for the Expected Parrot API.
    """

    def __init__(self, api_key: str = None, url: str = None) -> None:
        """
        Initialize the client.
        - Provide an API key directly, or through an env variable.
        - Provide a URL directly, or use the default one.
        """
        self.api_key = api_key or os.getenv("EXPECTED_PARROT_API_KEY")
        self.url = url or CONFIG.EXPECTED_PARROT_URL
        self._edsl_version = edsl.__version__

    ################
    # BASIC METHODS
    ################
    @property
    def headers(self) -> dict:
        """
        Return the headers for the request.
        """
        headers = {}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        else:
            headers["Authorization"] = f"Bearer None"
        return headers

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Send a request to the server and return the response.
        """
        url = f"{self.url}/{uri}"
        try:
            method = method.upper()
            if method in ["GET", "DELETE"]:
                response = requests.request(
                    method, url, params=params, headers=self.headers
                )
            elif method in ["POST", "PATCH"]:
                response = requests.request(
                    method, url, params=params, json=payload, headers=self.headers
                )
            else:
                raise Exception(f"Invalid {method=}.")
        except requests.ConnectionError:
            raise requests.ConnectionError("Could not connect to the server.")

        return response

    def _resolve_server_response(self, response: requests.Response) -> None:
        """
        Check the response from the server and raise errors as appropriate.
        """
        if response.status_code >= 400:
            message = response.json().get("detail")
            if "Authorization" in message:
                print(message)
                message = "Please provide an Expected Parrot API key."
            raise Exception(message)

    def _json_handle_none(self, value: Any) -> Any:
        """
        Handle None values during JSON serialization.
        - Return "null" if the value is None. Otherwise, don't return anything.
        """
        if value is None:
            return "null"

    @property
    def edsl_settings(self) -> dict:
        """
        Retrieve and return the EDSL settings stored on Coop.
        """
        response = self._send_server_request(uri="api/v0/edsl-settings", method="GET")
        self._resolve_server_response(response)
        return response.json()

    ################
    # Objects
    ################

    # TODO: add URL to get and get_all methods

    def create(
        self,
        object: EDSLObject,
        visibility: VisibilityType = "unlisted",
    ) -> dict:
        """
        Create an EDSL object in the Coop server.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(object)
        object_page = ObjectRegistry.get_object_page_by_object_type(object_type)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="POST",
            payload={
                "object_type": object_type,
                "json_string": json.dumps(
                    object.to_dict(),
                    default=self._json_handle_none,
                ),
                "visibility": visibility,
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        response_json = response.json()
        return {
            "uuid": response_json.get("uuid"),
            "version": self._edsl_version,
            "visibility": response_json.get("visibility"),
            "url": f"{self.url}/explore/{object_page}/{response_json.get('uuid')}",
        }

    def get(
        self,
        object_type: ObjectType = None,
        uuid: Union[str, UUID] = None,
        url: str = None,
    ) -> EDSLObject:
        """
        Retrieve an EDSL object either by object type & UUID, or by its url.
        - The object has to belong to the user or not be private.
        - Returns the initialized object class instance.

        :param object_type: the type of object to retrieve.
        :param uuid: the uuid of the object either in str or UUID format.
        :param url: the url of the object.
        """
        if url:
            object_type = url.split("/")[-2]
            uuid = url.split("/")[-1]
        elif not object_type and not uuid:
            raise Exception("Provide either object_type & UUID, or a url.")
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="GET",
            params={"type": object_type, "uuid": uuid},
        )
        self._resolve_server_response(response)
        json_string = response.json().get("json_string")
        return edsl_class.from_dict(json.loads(json_string))

    def _get_base(
        self,
        cls: EDSLObject,
        uuid: Union[str, UUID],
    ) -> EDSLObject:
        """
        Used by the Base class to offer a get functionality.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(cls)
        return self.get(object_type, uuid)

    def get_all(self, object_type: ObjectType) -> list[EDSLObject]:
        """
        Retrieve all objects of a certain type associated with the user.
        """
        edsl_class = ObjectRegistry.object_type_to_edsl_class.get(object_type)
        object_page = ObjectRegistry.get_object_page_by_object_type(object_type)
        response = self._send_server_request(
            uri=f"api/v0/objects",
            method="GET",
            params={"type": object_type},
        )
        self._resolve_server_response(response)
        objects = [
            {
                "object": edsl_class.from_dict(json.loads(o.get("json_string"))),
                "uuid": o.get("uuid"),
                "version": o.get("version"),
                "visibility": o.get("visibility"),
                "url": f"{self.url}/explore/{object_page}/{o.get('uuid')}",
            }
            for o in response.json()
        ]
        return objects

    def delete(self, object_type: ObjectType, uuid: Union[str, UUID]) -> dict:
        """
        Delete an object from the server.
        """
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="DELETE",
            params={"type": object_type, "uuid": uuid},
        )
        self._resolve_server_response(response)
        return response.json()

    def _delete_base(
        self,
        cls: EDSLObject,
        uuid: Union[str, UUID],
    ) -> dict:
        """
        Used by the Base class to offer a delete functionality.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(cls)
        return self.delete(object_type, uuid)

    def patch(
        self,
        object_type: ObjectType,
        uuid: Union[str, UUID],
        visibility: VisibilityType,
    ) -> dict:
        """
        Change the attributes of an uploaded object
        - Only supports visibility for now
        """
        response = self._send_server_request(
            uri=f"api/v0/object",
            method="PATCH",
            params={"type": object_type, "uuid": uuid},
            payload={"visibility": visibility},
        )
        self._resolve_server_response(response)
        return response.json()

    def _patch_base(
        self,
        cls: EDSLObject,
        uuid: Union[str, UUID],
        visibility: VisibilityType,
    ) -> dict:
        """
        Used by the Base class to offer a patch functionality.
        """
        object_type = ObjectRegistry.get_object_type_by_edsl_class(cls)
        return self.patch(object_type, uuid, visibility)

    ################
    # Remote Cache
    ################
    def remote_cache_create(
        self,
        cache_entry: CacheEntry,
        visibility: VisibilityType = "private",
    ) -> dict:
        """
        Create a single remote cache entry.
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache",
            method="POST",
            payload={
                "json_string": json.dumps(cache_entry.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    def remote_cache_create_many(
        self,
        cache_entries: list[CacheEntry],
        visibility: VisibilityType = "private",
    ) -> dict:
        """
        Create many remote cache entries.
        """
        payload = [
            {
                "json_string": json.dumps(c.to_dict()),
                "version": self._edsl_version,
                "visibility": visibility,
            }
            for c in cache_entries
        ]
        response = self._send_server_request(
            uri="api/v0/remote-cache/many",
            method="POST",
            payload=payload,
        )
        self._resolve_server_response(response)
        return response.json()

    def remote_cache_get(
        self,
        exclude_keys: Optional[list[str]] = None,
    ) -> list[CacheEntry]:
        """
        Get all remote cache entries.
        - optional exclude_keys: exclude CacheEntry objects with these keys.
        """
        if exclude_keys is None:
            exclude_keys = []
        response = self._send_server_request(
            uri="api/v0/remote-cache/get-many",
            method="POST",
            payload={"keys": exclude_keys},
        )
        self._resolve_server_response(response)
        return [
            CacheEntry.from_dict(json.loads(v.get("json_string")))
            for v in response.json()
        ]

    def remote_cache_clear(self) -> dict:
        """
        Clear all remote cache entries.
        """
        response = self._send_server_request(
            uri="api/v0/remote-cache/delete-all",
            method="DELETE",
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # Remote Inference
    ################
    def remote_inference_get(self, job_uuid: str) -> dict:
        """
        Get the results of a remote inference job.
        """
        response = self._send_server_request(
            uri="api/v0/remote-inference",
            method="GET",
            params={"uuid": job_uuid},
        )
        self._resolve_server_response(response)
        data = response.json()
        return {
            "jobs_uuid": data.get("jobs_uuid"),
            "results_uuid": data.get("results_uuid"),
            "results_url": "TO BE ADDED",
            "status": data.get("status"),
            "reason": data.get("reason"),
            "price": data.get("price"),
            "version": data.get("version"),
        }

    ################
    # Remote Errors
    ################
    def error_create(self, error_data: str) -> dict:
        """
        Send an error message to the server.
        """
        response = self._send_server_request(
            uri="api/v0/errors",
            method="POST",
            payload={
                "json_string": json.dumps(error_data),
                "version": self._edsl_version,
            },
        )
        self._resolve_server_response(response)
        return response.json()

    ################
    # DUNDER METHODS
    ################
    def __repr__(self):
        """Return a string representation of the client."""
        return f"Client(api_key='{self.api_key}', url='{self.url}')"

    ################
    # EXPERIMENTAL
    ################
    async def remote_async_execute_model_call(
        self, model_dict: dict, user_prompt: str, system_prompt: str
    ) -> dict:
        url = self.url + "/inference/"
        # print("Now using url: ", url)
        data = {
            "model_dict": model_dict,
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }
        # Use aiohttp to send a POST request asynchronously
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=data) as response:
                response_data = await response.json()
        return response_data

    def web(
        self,
        survey: dict,
        platform: Literal[
            "google_forms", "lime_survey", "survey_monkey"
        ] = "lime_survey",
        email=None,
    ):
        url = f"{self.url}/api/v0/export_to_{platform}"
        if email:
            data = {"json_string": json.dumps({"survey": survey, "email": email})}
        else:
            data = {"json_string": json.dumps({"survey": survey, "email": ""})}

        response_json = requests.post(url, headers=self.headers, data=json.dumps(data))

        return response_json


if __name__ == "__main__":
    from edsl.coop import Coop

    # init
    API_KEY = "b"
    coop = Coop(api_key=API_KEY)
    # basics
    coop
    coop.edsl_settings

    ##############
    # A. Objects
    ##############
    from uuid import uuid4
    from edsl import (
        Agent,
        AgentList,
        Cache,
        Jobs,
        QuestionMultipleChoice,
        Results,
        Scenario,
        ScenarioList,
        Survey,
    )

    OBJECTS = [
        ("agent", Agent),
        ("agent_list", AgentList),
        ("cache", Cache),
        ("job", Jobs),
        ("question", QuestionMultipleChoice),
        ("results", Results),
        ("scenario", Scenario),
        ("scenario_list", ScenarioList),
        ("survey", Survey),
    ]

    for object_type, cls in OBJECTS:
        print(f"Testing {object_type} objects")
        # 1. Delete existing objects
        existing_objects = coop.get_all(object_type)
        for item in existing_objects:
            coop.delete(object_type=object_type, uuid=item.get("uuid"))
        # 2. Create new objects
        example = cls.example()
        response_1 = coop.create(example)
        response_2 = coop.create(cls.example(), visibility="private")
        response_3 = coop.create(cls.example(), visibility="public")
        response_4 = coop.create(cls.example(), visibility="unlisted")
        # 3. Retrieve all objects
        objects = coop.get_all(object_type)
        assert len(objects) == 4
        # 4. Try to retrieve an item that does not exist
        try:
            coop.get(object_type=object_type, uuid=uuid4())
        except Exception as e:
            print(e)
        # 5. Try to retrieve all test objects by their uuids
        for response in [response_1, response_2, response_3, response_4]:
            coop.get(object_type=object_type, uuid=response.get("uuid"))
        # 6. Change visibility of all objects
        for item in objects:
            coop.patch(
                object_type=object_type, uuid=item.get("uuid"), visibility="private"
            )
        # 7. Delete all objects
        for item in objects:
            coop.delete(object_type=object_type, uuid=item.get("uuid"))
        assert len(coop.get_all(object_type)) == 0

    response = QuestionMultipleChoice.example().push()
    QuestionMultipleChoice.pull(response.get("uuid"))
    coop.patch(object_type="question", uuid=response.get("uuid"), visibility="public")
    coop.delete(object_type="question", uuid=response.get("uuid"))

    ##############
    # B. Remote Cache
    ##############
    from edsl.data.CacheEntry import CacheEntry

    # clear
    coop.remote_cache_clear()
    assert coop.remote_cache_get() == []
    # create one remote cache entry
    cache_entry = CacheEntry.example()
    cache_entry.to_dict()
    coop.remote_cache_create(cache_entry)
    # create many remote cache entries
    cache_entries = [CacheEntry.example(randomize=True) for _ in range(10)]
    coop.remote_cache_create_many(cache_entries)
    # get all remote cache entries
    coop.remote_cache_get()
    coop.remote_cache_get(exclude_keys=[])
    coop.remote_cache_get(exclude_keys=["a"])
    exclude_keys = [cache_entry.key for cache_entry in cache_entries]
    coop.remote_cache_get(exclude_keys)
    # clear
    coop.remote_cache_clear()
    coop.remote_cache_get()

    ##############
    # C. Remote Inference
    ##############
    from edsl.jobs import Jobs

    # check jobs on server (should be an empty list)
    coop.get_all("job")
    for job in coop.get_all("job"):
        coop.delete(object_type="job", uuid=job.get("uuid"))
    # post a job
    response = coop.create(Jobs.example())
    # get job and results
    coop.remote_inference_get(response.get("uuid"))
    coop.get(
        object_type="results",
        uuid=coop.remote_inference_get(response.get("uuid")).get("results_uuid"),
    )

    ##############
    # D. Errors
    ##############
    from edsl import Coop

    coop = Coop()
    coop.api_key = "a"
    coop.error_create({"something": "This is an error message"})
    coop.api_key = None
    coop.error_create({"something": "This is an error message"})



import asyncio
from typing import Callable, Union, List
from collections import UserList, UserDict


from edsl.jobs.buckets import ModelBuckets
from edsl.exceptions import InterviewErrorPriorTaskCanceled

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tasks.task_status_enum import TaskStatus, TaskStatusDescriptor

# from edsl.jobs.tasks.task_management import TokensUsed
from edsl.jobs.tasks.TaskStatusLog import TaskStatusLog
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage
from edsl.jobs.tokens.TokenUsage import TokenUsage
from edsl.jobs.Answers import Answers

from edsl.questions.QuestionBase import QuestionBase


class TokensUsed(UserDict):
    """ "Container for tokens used by a task."""

    def __init__(self, cached_tokens, new_tokens):
        d = {"cached_tokens": cached_tokens, "new_tokens": new_tokens}
        super().__init__(d)


class QuestionTaskCreator(UserList):
    """Class to create and manage a single question and its dependencies.
    The class is an instance of a UserList of tasks that must be completed before the focal task can be run.

    It is a UserList with all the tasks that must be completed before the focal task can be run.
    The focal task is the question that we are interested in answering.
    """

    task_status = TaskStatusDescriptor()

    def __init__(
        self,
        *,
        question: QuestionBase,
        answer_question_func: Callable,
        model_buckets: ModelBuckets,
        token_estimator: Union[Callable, None] = None,
        iteration: int = 0,
    ):
        """Initialize the QuestionTaskCreator instance.

        :param question: the question that we are interested in answering.
        :param answer_question_func: the function that will answer the question.
        :param model_buckets: the bucket collection that contains the requests and tokens buckets which control the rate of API calls and token usage.
        :param token_estimator: a function that estimates the number of tokens required to answer the question.
        :param iteration: the iteration number of the question.

        """
        super().__init__([])
        self.answer_question_func = answer_question_func
        self.question = question
        self.iteration = iteration

        self.model_buckets = model_buckets
        self.requests_bucket = self.model_buckets.requests_bucket
        self.tokens_bucket = self.model_buckets.tokens_bucket
        self.status_log = TaskStatusLog()

        def fake_token_estimator(question):
            return 1

        self.token_estimator = token_estimator or fake_token_estimator

        # Assume that the task is *not* from the cache until we know otherwise; the _run_focal_task might flip this bit later.
        self.from_cache = False

        self.cached_token_usage = TokenUsage(from_cache=True)
        self.new_token_usage = TokenUsage(from_cache=False)
        self.task_status = TaskStatus.NOT_STARTED

    def add_dependency(self, task: asyncio.Task) -> None:
        """Adds a task dependency to the list of dependencies.

        >>> qt1 = QuestionTaskCreator.example()
        >>> qt2 = QuestionTaskCreator.example()
        >>> qt2.add_dependency(qt1)
        >>> len(qt2)
        1
        """
        self.append(task)

    def generate_task(self, debug: bool) -> asyncio.Task:
        """Create a task that depends on the passed-in dependencies."""
        task = asyncio.create_task(
            self._run_task_async(debug), name=self.question.question_name
        )
        task.depends_on = [t.get_name() for t in self]
        return task

    def estimated_tokens(self) -> int:
        """Estimates the number of tokens that will be required to run the focal task."""
        return self.token_estimator(self.question)

    def token_usage(self) -> TokensUsed:
        """Returns the token usage for the task.

        >>> qt = QuestionTaskCreator.example()
        >>> answers = asyncio.run(qt._run_focal_task(debug=False))
        >>> qt.token_usage()
        {'cached_tokens': TokenUsage(from_cache=True, prompt_tokens=0, completion_tokens=0), 'new_tokens': TokenUsage(from_cache=False, prompt_tokens=0, completion_tokens=0)}
        """
        return TokensUsed(
            cached_tokens=self.cached_token_usage, new_tokens=self.new_token_usage
        )

    async def _run_focal_task(self, debug: bool) -> Answers:
        """Run the focal task i.e., the question that we are interested in answering.

        It is only called after all the dependency tasks are completed.

        >>> qt = QuestionTaskCreator.example()
        >>> answers = asyncio.run(qt._run_focal_task(debug=False))
        >>> answers["answer"]
        'Yo!'
        """

        requested_tokens = self.estimated_tokens()
        if (estimated_wait_time := self.tokens_bucket.wait_time(requested_tokens)) > 0:
            self.task_status = TaskStatus.WAITING_FOR_TOKEN_CAPACITY

        await self.tokens_bucket.get_tokens(requested_tokens)

        if (estimated_wait_time := self.requests_bucket.wait_time(1)) > 0:
            self.waiting = True
            self.task_status = TaskStatus.WAITING_FOR_REQUEST_CAPACITY

        await self.requests_bucket.get_tokens(1)

        self.task_status = TaskStatus.API_CALL_IN_PROGRESS
        try:
            results = await self.answer_question_func(
                question=self.question, debug=debug, task=None  # self
            )
            self.task_status = TaskStatus.SUCCESS
        except Exception as e:
            self.task_status = TaskStatus.FAILED
            raise e

        if "cached_response" in results:
            if results["cached_response"]:
                # Gives back the tokens b/c the API was not called.
                self.tokens_bucket.add_tokens(requested_tokens)
                self.requests_bucket.add_tokens(1)
                self.from_cache = True

        tracker = self.cached_token_usage if self.from_cache else self.new_token_usage

        # TODO: This is hacky. The 'func' call should return an object that definitely has a 'usage' key.
        usage = results.get("usage", {"prompt_tokens": 0, "completion_tokens": 0})
        prompt_tokens = usage.get("prompt_tokens", 0)
        completion_tokens = usage.get("completion_tokens", 0)
        tracker.add_tokens(
            prompt_tokens=prompt_tokens, completion_tokens=completion_tokens
        )

        return results

    @classmethod
    def example(cls):
        """Return an example instance of the class."""
        from edsl import QuestionFreeText
        from edsl.jobs.buckets.ModelBuckets import ModelBuckets

        m = ModelBuckets.infinity_bucket()

        async def answer_question_func(question, debug, task):
            return {"answer": "Yo!"}

        return cls(
            question=QuestionFreeText.example(),
            answer_question_func=answer_question_func,
            model_buckets=m,
            token_estimator=None,
            iteration=0,
        )

    async def _run_task_async(self, debug) -> None:
        """Run the task asynchronously, awaiting the tasks that must be completed before this one can be run.

        >>> qt1 = QuestionTaskCreator.example()
        >>> qt2 = QuestionTaskCreator.example()
        >>> qt2.add_dependency(qt1)

        The method follows these steps:
            1. Set the task_status to TaskStatus.WAITING_FOR_DEPENDENCIES, indicating that the task is waiting for its dependencies to complete.
            2. Await asyncio.gather(*self, return_exceptions=True) to run all the dependent tasks concurrently.

            - the return_exceptions=True flag ensures that the task does not raise an exception if any of the dependencies fail.

            3. If any of the dependencies raise an exception:
            - If it is a CancelledError, set the current task's task_status to TaskStatus.CANCELLED, and re-raise the CancelledError,
                terminating the execution of the current task.
            - If it is any other exception, set the task_status to TaskStatus.PARENT_FAILED, and raise a custom exception
                InterviewErrorPriorTaskCanceled with the original exception as the cause, terminating the execution of the current task.
            4. If all the dependencies complete successfully without raising any exceptions, the code reaches the else block.
            5. In the else block, run the focal task (self._run_focal_task(debug)).

            If any of the dependencies fail (raise an exception), the focal task will not run. The execution will be terminated,
            and an exception will be raised to indicate the failure of the dependencies.

            The focal task (self._run_focal_task(debug)) is only executed if all the dependencies complete successfully.

            Args:
                debug: A boolean value indicating whether to run the task in debug mode.

            Returns:
                None
        """
        try:
            self.task_status = TaskStatus.WAITING_FOR_DEPENDENCIES
            # If this were set to 'return_exceptions=False', then the first exception would be raised immediately.
            # and it would cancel all the other tasks. This is not the behavior we want.

            gather_results = await asyncio.gather(*self, return_exceptions=True)

            for result in gather_results:
                if isinstance(result, Exception):
                    raise result

            return await self._run_focal_task(debug)

        except asyncio.CancelledError:
            self.task_status = TaskStatus.CANCELLED
            raise
        except Exception as e:
            # one of the dependencies failed
            self.task_status = TaskStatus.PARENT_FAILED
            # turns the parent exception into a custom exception so the task gets canceled but this InterviewErrorPriorTaskCanceled exception
            raise InterviewErrorPriorTaskCanceled(
                f"Required tasks failed for {self.question.question_name}"
            ) from e


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from collections import UserDict
from typing import List, Dict, Optional
import textwrap
import re

from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.QuestionFreeText import QuestionFreeText

from edsl.conjure.utilities import convert_value, Missing


class KeyValidator:
    """A class to represent a key validator.

    >>> k = KeyValidator()
    >>> k.validate_key("asdf")
    True
    >>> k.validate_key("ASDF")
    False

    """

    def __set_name__(self, owner, name):
        self.name = name

    def validate_key(self, key):
        if not isinstance(key, str):
            # "Key must be a string"
            return False
        if key.lower() != key:
            # "Key must be lowercase"
            return False
        if not key.isidentifier() or not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", key):
            # raise ValueError("Key must be a valid Python identifier")
            return False
        return True


class ReplacementFinder:
    """This class finds a replacement name for a bad question name.

    >>> r = ReplacementFinder(lookup_dict = {'Poop ': 'poop'})
    >>> r('Poop ')
    'poop'

    """

    def __init__(self, lookup_dict: Optional[dict] = None):
        if lookup_dict is None:
            lookup_dict = {}

        self.lookup_dict = lookup_dict

    def __call__(self, bad_question_name):
        """Finds a replacement name for a bad question name.
        TODO: We should add a check to see if the new name is already in use.
        """
        if bad_question_name in self.lookup_dict:
            return self.lookup_dict[bad_question_name]

        q = QuestionFreeText(
            question_text=f"""We have a survey with a question name: {bad_question_name}. 
            The question name is not a valid Python identifier.
            We need a valid Python identifier to use as a column name in a dataframe.
            What would be a better name for this question?
            Shorter is better.
            Just return the proposed identifier with no other text.
            """,
            question_name="identifier",
        )
        new_identifer = q.run().select("identifier").first().lower()
        self.lookup_dict[bad_question_name] = new_identifer
        return new_identifer

    def __repr__(self):
        return f"ReplacementFinder({self.lookup_dict})"

    def to_json(self):
        return self.lookup_dict

    @classmethod
    def from_json(cls, json_dict):
        return cls(json_dict)


get_replacement_name = ReplacementFinder({})


class CustomDict(UserDict):
    """
    This class is a dictionary that only allows lowercase keys that are valid Python identifiers.
    If a key is not a valid Python identifier, it will be replaced with a valid Python identifier.

    >>> d = CustomDict()
    >>> d = CustomDict({"7asdf": 123, "FAMILY": 12})
    >>> d
    {'q7asdf': 123, 'family': 12}
    """

    key_validator = KeyValidator()

    def __init__(self, data=None, verbose=False):
        super().__init__()
        self.verbose = verbose
        if data:
            for key, value in data.items():
                self[key] = value

    def __setitem__(self, key, value):
        if key != key.lower():
            key = key.lower()
        while not self.key_validator.validate_key(key):
            if self.verbose:
                print(f"Column heading incapable of being a key: {key}")
            if key in get_replacement_name.lookup_dict:
                key = get_replacement_name.lookup_dict[key]
            else:
                key = get_replacement_name(key)
            if self.verbose:
                print(f"New key: {key}")
        super().__setitem__(key, value)


class RawResponseColumn:
    """A class to represent a raw responses from a survey we are parsing.

    If the dataset was square, we would think of this as a column in a dataframe.

    """

    def __init__(
        self,
        question_name: str,
        raw_responses: List[str],
        answer_codebook: Dict[str, str],
        question_text: str,
    ):
        """
        :param question_name: The name of the question.
        :param raw_responses: A list of responses to the question.
        :param answer_codebook: A dictionary mapping the raw responses to the actual responses.
        :param question_text: The text of the question.

        >>> r = RawResponseColumn(question_name="Q1", raw_responses=["1", "2", "3"], answer_codebook={"1": "Yes", "2": "No"}, question_text="Do you like ice cream?")
        >>> r.responses
        ['Yes', 'No', '3']
        >>> r.question_name
        'q1'

        >>> r = RawResponseColumn(question_name="Q1", raw_responses=["1", "2", "3"], answer_codebook={"1": "Yes", "2": "No"}, question_text="Do you like ice cream?")
        >>> r.inferred_question_type
        'multiple_choice'

        """
        d = CustomDict(
            {question_name: ""}
        )  # if the question_name is not a valid Python identifier, it will be replaced with a valid Python identifier.
        self.question_name = list(d.keys())[0]
        self.raw_responses = raw_responses
        self.answer_codebook = answer_codebook
        if question_text is None or len(question_text) == 0:
            self.question_text = "Question text not available"
        else:
            self.question_text = question_text

    @staticmethod
    def edsl_question_inference(question_text, responses):
        """Infer the question type from the responses.

        TODO: Not currently used. We should use this to infer the question type.

        >>> from edsl import QuestionFreeText
        >>> q = QuestionFreeText(question_text="Tell me about your childhood.", question_name = 'color')
        >>> r = RawResponseColumn.edsl_question_inference(q.question_text, ["Rather not say", "Happy!", "luge lessons and meat helmets"])
        >>> r
        'free_text'

        """
        from edsl.questions import QuestionMultipleChoice

        q = QuestionMultipleChoice(
            question_text="""We have a survey question and we are trying to infer its type.
                               The question text is: '{{question_text}}'.                                   
                               The first few responses are: '{{responses}}'.
                                """,
            question_name="infer_question_type",
            question_options=[
                "budget",
                "checkbox",
                "extract",
                "free_text",
                "likert_five",
                "linear_scale",
                "list",
                "multiple_choice",
                "numerical",
                "rank",
                "top_k",
                "yes_no",
            ],
        )
        response = (
            q.to_survey()(question_text=question_text, responses=responses)
            .select("infer_question_type")
            .first()
        )
        return response

    @property
    def responses(self):
        """Returns the responses, with the answer codebook applied."""
        if hasattr(self, "answer_codebook") is None:
            converted_responses = [convert_value(x) for x in self.raw_responses]
            return converted_responses
        else:
            return [self.answer_codebook.get(x, x) for x in self.raw_responses]

    @property
    def inferred_question_type(self):
        "Tries to infer the type of question from the responses and other information"
        max_items = 15
        options = list(self.unique_responses.keys())
        if len(options) > max_items or len(options) <= 1:
            return "free_text"
        else:
            cv = [convert_value(o) for o in options]
            return "multiple_choice"

    def get_ordering(self, options_list):
        """Returns a multiple choice question with the options in the correct order.
        For example, if the options are ["<10", "10-20", "30+", "20-30"], the question would be:
        """
        from edsl.questions.QuestionList import QuestionList

        q = QuestionList(
            question_text=textwrap.dedent(
                """\
            We have a survey question with the following options: '{{options_list}}'.
            The options might be out of order. Please put them in the correct order.
            If there is not natural order, just put then in order they were presented.
            """
            ),
            question_name="ordering",
        )

        return q.to_survey()(options_list=options_list).select("ordering").first()

    def to_question(self) -> QuestionBase:
        """Returns a Question object."""
        d = {}
        d["question_name"] = self.question_name
        d["question_text"] = self.question_text
        d["question_type"] = self.inferred_question_type
        if d["question_type"] == "multiple_choice":
            d["question_options"] = self.get_ordering(list(self.unique_responses))
        return QuestionBase.from_dict(d)

    @property
    def unique_responses(self) -> set:
        """Returns the unique responses for a given question; useful to build up mulitple choice questions.

        >>> r = RawResponseColumn(question_name="Q1", raw_responses=["1", "2", "3"], answer_codebook={"1": "Yes", "2": "No"}, question_text="Do you like ice cream?")
        >>> r.unique_responses
        {'Yes', 'No', '3'}
        """
        from collections import defaultdict

        s = defaultdict(int)
        for response in self.responses:
            if isinstance(response, Missing):
                continue
            else:
                s[str(response)] += 1
        return s

    def response_category(self):
        """Returns the inferred category of the question."""
        from edsl.questions import QuestionMultipleChoice

        q = QuestionMultipleChoice(
            question_text=textwrap.dedent(
                f"""\
            I have data from a survey. One of the column labels: {self.question_text}. 
            I want to know what the data represents."""
            ),
            question_options=[
                "Something asked of the respondent that interested the survey company",
                "A fixed characteristic of the respondent, like gender or age",
                "Measured by survey company like an internal code or start of the interview",
            ],
            question_name="opinion_or_characteristic",
        )
        result = q.run().select("opinion_or_characteristic").first()
        return result

    def __repr__(self):
        raw_response_display = self.raw_responses[0 : min(len(self.responses), 5)]
        if len(self.responses) > 5:
            raw_response_display.append("...")

        raw_response_display = self.responses[0 : min(len(self.responses), 5)]
        if len(self.responses) > 5:
            raw_response_display.append("...")

        return f"""RawResponseColumn(question_name="{self.question_name}", question_text="{self.question_text}", raw_responses={raw_response_display}, responses={raw_response_display}, unqiue_responses={self.unique_responses}, answer_codebook={self.answer_codebook})"""


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # d = CustomDict()
    # d = CustomDict({"7asdf": 123, "FAMILY": 12})
    # d["a"] = 123
    # d["#_family_members"] = 4

    # d["FAMILY_MEMBERS"] = 12
    # d["0x1389"] = 123
    # print(d)
    # r = RawResponseColumn(
    #     question_name="_x family MeMbers",
    #     raw_responses=["1", "2", "3"],
    #     question_text="fake",
    #     answer_codebook={},
    # )

    # get_replacement_name



class ResultsErrors(Exception):
    pass


class ResultsBadMutationstringError(ResultsErrors):
    pass


class ResultsColumnNotFoundError(ResultsErrors):
    pass


class ResultsInvalidNameError(ResultsErrors):
    pass


class ResultsMutateError(ResultsErrors):
    pass


class ResultsFilterError(ResultsErrors):
    pass



import enum


class InterviewStatus(enum.Enum):
    "These are the possible states an interview can be in."
    NOT_STARTED = enum.auto()
    SUCCESS = enum.auto()
    WAITING_FOR_RESOURCES = enum.auto()
    FAILED = enum.auto()
